import os.path

from depscan.lib.csaf import (
    CsafOccurence,
    add_vulnerabilities,
    cleanup_dict,
    cleanup_list,
    format_references,
    get_product_status,
    get_ref_summary,
    import_csaf_toml,
    import_root_component,
    parse_cvss,
    parse_cwe,
    parse_revision_history,
    parse_toml,
    verify_components_present,
    version_helper,
)


def test_parse_revision_history():
    # add revision entry w/existing entries when final
    tracking = {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "status": "final",
        "version": "2",
        "revision_history": [
            {
                "date": "2023-10-02T23:50:07",
                "number": "2",
                "summary": "Update",
            },
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            },
        ],
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "revision_history": [
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            },
            {
                "date": "2023-10-02T23:50:07",
                "number": "2",
                "summary": "Update",
            },
            {"date": "2023-10-03T00:21:34", "number": "3", "summary": "Update"},
        ],
        "status": "final",
        "version": "3",
    }

    # add revision entry w/no existing entries when final
    tracking = {
        "current_release_date": "2022-09-22T20:54:06.186927",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06.186927",
        "status": "final",
        "version": "",
        "revision_history": [],
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2022-09-22T20:54:06",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "revision_history": [
            {"date": "2022-09-22T20:54:06", "number": "1", "summary": "Initial"}
        ],
        "status": "final",
        "version": "1",
    }

    # do not add when status is not final
    tracking = {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "status": "draft",
        "version": "2",
        "revision_history": [
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            }
        ],
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "status": "draft",
        "version": "2",
        "revision_history": [
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            }
        ],
    }

    # deal with a revision history inconsistent with the version number
    tracking = {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "status": "final",
        "version": "5",
        "revision_history": [
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            }
        ],
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2023-10-03T00:21:34",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "status": "final",
        "version": "2",
        "revision_history": [
            {
                "date": "2022-09-22T20:54:06",
                "number": "1",
                "summary": "Initial",
            },
            {
                "date": "2023-10-03T00:21:34",
                "number": "2",
                "summary": "Update",
            },
        ],
    }

    # deal with a missing revision history
    tracking = {
        "current_release_date": "2022-09-22T20:54:06.186927",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06.186927",
        "status": "final",
        "version": "",
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2022-09-22T20:54:06",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "revision_history": [
            {"date": "2022-09-22T20:54:06", "number": "1", "summary": "Initial"}
        ],
        "status": "final",
        "version": "1",
    }

    # deal with a NoneType revision history
    tracking = {
        "current_release_date": "2022-09-22T20:54:06.186927",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06.186927",
        "status": "final",
        "version": "",
        "revision_history": None,
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2022-09-22T20:54:06",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "revision_history": [
            {"date": "2022-09-22T20:54:06", "number": "1", "summary": "Initial"}
        ],
        "status": "final",
        "version": "1",
    }

    # update initial release date when adding initial release
    tracking = {
        "current_release_date": "2022-09-22T20:54:06.186927",
        "id": "ID",
        "initial_release_date": "2022-08-22T20:54:06.186927",
        "status": "final",
        "version": "1",
        "revision_history": [],
    }
    assert parse_revision_history(tracking) == {
        "current_release_date": "2022-09-22T20:54:06",
        "id": "ID",
        "initial_release_date": "2022-09-22T20:54:06",
        "revision_history": [
            {"date": "2022-09-22T20:54:06", "number": "1", "summary": "Initial"}
        ],
        "status": "final",
        "version": "1",
    }


def test_cleanup_list():
    assert cleanup_list([{}]) == []
    assert cleanup_list([{"a": "a", "b": "b", "c": ""}]) == [
        {"a": "a", "b": "b"}
    ]
    assert cleanup_list(["test", None]) == ["test"]


def test_cleanup_dict():
    assert cleanup_dict({"test": {"a": []}}) == {}
    assert cleanup_dict({"test": ""}) == {}
    assert cleanup_dict({"test": "", "test2": "test2"}) == {"test2": "test2"}
    assert cleanup_dict({"a": "a", "b": "b", "c": ""}) == {"a": "a", "b": "b"}
    assert cleanup_dict({"dict": ["test", None]}) == {"dict": ["test"]}


def test_get_ref_summary():
    url = "https://nvd.nist.gov/vuln/detail/cve-2021-1234"
    assert get_ref_summary(url) == "CVE Record"
    url = "https://github.com/user/repo/security/advisories/GHSA-1234-1234-1234"
    assert get_ref_summary(url) == "Advisory"
    url = "https://github.com/user/repo/pull/123"
    assert get_ref_summary(url) == "GitHub Pull Request"
    url = "https://github.com/user/repo/commit/123"
    assert get_ref_summary(url) == "GitHub Commit"
    url = ""
    assert get_ref_summary(url) == "Other"
    url = "https://example.com"
    assert get_ref_summary(url) == "Other"
    url = "https://github.com/user/repo/release"
    assert get_ref_summary(url) == "GitHub Repository Release"
    url = "https://github.com/user/repo"
    assert get_ref_summary(url) == "GitHub Repository"
    url = "https://access.redhat.com/security/cve/CVE-2023-26136"
    assert get_ref_summary(url) == "CVE Record"
    url = "https://access.redhat.com/errata/RHSA-2023:5484"
    assert get_ref_summary(url) == "Advisory"
    url = "https://bugzilla.redhat.com/show_bug.cgi?id=2224245"
    assert get_ref_summary(url) == "Bugzilla"


def test_format_references():
    ref = [
        "https://access.redhat.com/errata/RHSA-2023:5484",
        "https://bugzilla.redhat.com/show_bug.cgi?id=2224245",
        "https://nvd.nist.gov/vuln/detail/cve-2021-1234",
        "https://github.com/advisories/GHSA-1234-1234-1234",
        "https://github.com/user/repo/security/advisories/GHSA-5432-5432-5432",
        "https://github.com/user/repo/pull/123",
        "https://github.com/user/repo/commit/123",
        "https://example.com",
        "https://github.com/user/repo/release",
        "https://github.com/user/repo",
        "https://bugzilla.redhat.com/show_bug.cgi?id=cve-2021-1234",
        "https://github.com/FasterXML/jackson-databind/issues/2816",
        "https://sec.cloudapps.cisco.com/security/center/content"
        "/CiscoSecurityAdvisory/cisco-sa-apache-log4j-qRuKNEbd",
        "https://bitbucket.org/snakeyaml/snakeyaml/issues/525",
        "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=47027",
    ]
    [ids, refs] = format_references(ref)
    # For consistency in tests
    ids = sorted(ids, key=lambda x: x["text"])
    refs = sorted(refs, key=lambda x: x["url"])
    assert ids == [
        {"system_name": "Red Hat Bugzilla ID", "text": "2224245"},
        {
            "system_name": "GitHub Issue [FasterXML/jackson-databind]",
            "text": "2816",
        },
        {"system_name": "Chromium Issue [oss-fuzz]", "text": "47027"},
        {"system_name": "Bitbucket Issue [snakeyaml/snakeyaml]", "text": "525"},
        {"system_name": "GitHub Advisory", "text": "GHSA-1234-1234-1234"},
        {"system_name": "GitHub Advisory", "text": "GHSA-5432-5432-5432"},
        {"system_name": "Red Hat Advisory", "text": "RHSA-2023:5484"},
        {
            "system_name": "Cisco Advisory",
            "text": "cisco-sa-apache-log4j-qRuKNEbd",
        },
        {"system_name": "Red Hat Bugzilla ID", "text": "cve-2021-1234"},
    ]
    assert refs == [
        {
            "summary": "Red Hat Advisory",
            "url": "https://access.redhat.com/errata/RHSA-2023:5484",
        },
        {
            "summary": "Bitbucket Issue",
            "url": "https://bitbucket.org/snakeyaml/snakeyaml/issues/525",
        },
        {
            "summary": "Chromium Issue",
            "url": "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=47027",
        },
        {
            "summary": "Red Hat Bugzilla",
            "url": "https://bugzilla.redhat.com/show_bug.cgi?id=2224245",
        },
        {
            "summary": "Red Hat Bugzilla",
            "url": "https://bugzilla.redhat.com/show_bug.cgi?id=cve-2021-1234",
        },
        {"summary": "Other", "url": "https://example.com"},
        {
            "summary": "GitHub Issue",
            "url": "https://github.com/FasterXML/jackson-databind/issues/2816",
        },
        {
            "summary": "GitHub Advisory",
            "url": "https://github.com/advisories/GHSA-1234-1234-1234",
        },
        {"summary": "GitHub Repository", "url": "https://github.com/user/repo"},
        {
            "summary": "GitHub Commit",
            "url": "https://github.com/user/repo/commit/123",
        },
        {
            "summary": "GitHub Pull Request",
            "url": "https://github.com/user/repo/pull/123",
        },
        {
            "summary": "GitHub Repository Release",
            "url": "https://github.com/user/repo/release",
        },
        {
            "summary": "GitHub Advisory",
            "url": "https://github.com/user/repo/security/advisories/GHSA"
            "-5432-5432-5432",
        },
        {
            "summary": "CVE Record",
            "url": "https://nvd.nist.gov/vuln/detail/cve-2021-1234",
        },
        {
            "summary": "Cisco Advisory",
            "url": "https://sec.cloudapps.cisco.com/security/center/content"
            "/CiscoSecurityAdvisory/cisco-sa-apache-log4j-qRuKNEbd",
        },
    ]


def test_parse_cwe():
    assert parse_cwe("['CWE-20', 'CWE-668']") == (
        {"id": "CWE-20", "name": "Improper Input Validation"},
        [
            {
                "title": "Additional CWE: CWE-668",
                "audience": "developers",
                "category": "other",
                "text": "Exposure of Resource to Wrong Sphere",
            }
        ],
    )
    assert parse_cwe("CWE-1333") == (
        {
            "id": "CWE-1333",
            "name": "Inefficient Regular Expression Complexity",
        },
        [],
    )
    assert parse_cwe("") == (None, [])
    assert parse_cwe("CWE-000") == (None, [])


def test_parse_toml():
    # If running tests using an IDE such as PyCharm, pytest may execute from
    # the test directory rather than the project root.
    if os.path.exists(os.path.join(os.getcwd(), "contrib/csaf.toml")):
        filepath = os.path.join(os.getcwd(), "contrib/csaf.toml")
    else:
        filepath = "../contrib/csaf.toml"
    metadata = import_csaf_toml(filepath)
    # We don't want a dynamically generated ID
    metadata["tracking"]["id"] = "1234"
    parsed_toml = parse_toml(metadata)
    assert parsed_toml["document"]["category"] == "csaf_vex"
    assert parsed_toml["document"]["notes"] == [
        {"audience": "", "category": "", "text": "", "title": ""}
    ]
    assert parsed_toml["document"]["publisher"] == {
        "category": "vendor",
        "contact_details": "vendor@mcvendorson.com",
        "name": "Vendor McVendorson",
        "namespace": "https://appthreat.com",
    }


def test_parse_cvss():
    # Test parsing
    res = {
        "cvss_v3": {
            "attack_complexity": "LOW",
            "attack_vector": "NETWORK",
            "availability_impact": "HIGH",
            "base_score": 7.5,
            "impact_score": 7.5,
            "confidentiality_impact": "NONE",
            "integrity_impact": "NONE",
            "privileges_required": "NONE",
            "scope": "UNCHANGED",
            "user_interaction": "NONE",
            "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H",
        },
        "severity": "HIGH",
        "id": "CVE-2023-37788",
    }
    assert parse_cvss(res) == {
        "baseScore": 7.5,
        "attackVector": "NETWORK",
        "privilegesRequired": "NONE",
        "userInteraction": "NONE",
        "scope": "UNCHANGED",
        "baseSeverity": "HIGH",
        "version": "3.1",
        "vectorString": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H",
    }
    res["cvss_v3"]["vector_string"] = "CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I"
    assert parse_cvss(res) == {
        "baseScore": 7.5,
        "attackVector": "NETWORK",
        "privilegesRequired": "NONE",
        "userInteraction": "NONE",
        "scope": "UNCHANGED",
        "baseSeverity": "HIGH",
        "version": "3.0",
        "vectorString": "CVSS:3.0/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I",
    }
    # Test no cvss_v3
    res = {
        "severity": "HIGH",
        "id": "CVE-2023-37788",
    }
    assert parse_cvss(res) is None
    res["cvss_v3"] = {}
    assert parse_cvss(res) is None
    # Test missing or pre-3.0 vector string
    res = {
        "cvss_v3": {
            "attack_complexity": "LOW",
            "attack_vector": "NETWORK",
            "availability_impact": "HIGH",
            "base_score": 7.5,
            "impact_score": 7.5,
            "confidentiality_impact": "NONE",
            "integrity_impact": "NONE",
            "privileges_required": "NONE",
            "scope": "UNCHANGED",
            "user_interaction": "NONE",
        },
        "severity": "HIGH",
        "id": "CVE-2023-37788",
    }
    assert parse_cvss(res) is None
    res["cvss_v3"]["vector_string"] = "CVSS:2.0/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I"
    assert parse_cvss(res) is None
    res["cvss_v3"]["vector_string"] = ""
    assert parse_cvss(res) is None
    res["cvss_v3"]["vector_string"] = None
    assert parse_cvss(res) is None
    # Test missing base score
    res = {
        "cvss_v3": {
            "attack_complexity": "LOW",
            "attack_vector": "NETWORK",
            "availability_impact": "HIGH",
            "impact_score": 7.5,
            "confidentiality_impact": "NONE",
            "integrity_impact": "NONE",
            "privileges_required": "NONE",
            "scope": "UNCHANGED",
            "user_interaction": "NONE",
            "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H",
        },
        "severity": "HIGH",
        "id": "CVE-2023-37788",
    }
    assert parse_cvss(res) is None
    res["cvss_v3"]["base_score"] = ""
    assert parse_cvss(res) is None
    res["cvss_v3"]["base_score"] = None
    assert parse_cvss(res) is None


def test_get_product_status():
    assert get_product_status(
        {
            "affected_location": {
                "cpe_uri": "cpe:2.3:a:npm:taffydb:*:*:*:*:*:*:*:*",
                "package": "taffydb",
                "version": "<=2.7.3",
            },
            "fixed_location": None,
        },
        "1089386|taffydb|2.6.2",
    ) == (
        "taffydb",
        {"known_affected": ["taffydb:<=2.7.3"]},
        "<=2.7.3",
        "taffydb",
    )

    assert get_product_status(
        {
            "affected_location": {
                "cpe_uri": "cpe:2.3:a:npm:taffydb:*:*:*:*:*:*:*:*",
                "package": "taffydb",
                "version": "<=2.7.3",
            },
            "fixed_location": None,
        },
        "1089386_32636283|taffygroup|taffydb|2.6.2",
    ) == (
        "taffydb",
        {"known_affected": ["taffydb:<=2.7.3"]},
        "<=2.7.3",
        "taffygroup/taffydb",
    )


def test_csaf_occurence():
    res = [
        {
            "id": "CVE-2019-10790",
            "problem_type": "['CWE-20', 'CWE-668']",
            "type": "npm",
            "severity": None,
            "cvss_score": "7.5",
            "cvss_v3": {
                "base_score": 7.5,
                "exploitability_score": 7.5,
                "impact_score": 7.5,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "REQUIRED",
                "scope": "UNCHANGED",
                "confidentiality_impact": "HIGH",
                "integrity_impact": "HIGH",
                "availability_impact": "HIGH",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:L",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:npm:taffydb:*:*:*:*:*:*:*:*",
                    "package": "taffydb",
                    "version": "<=2.7.3",
                },
                "fixed_location": None,
            },
            "short_description": "# TaffyDB can allow access to any data items "
            "in the DB\nTaffyDB allows attackers to forge "
            "adding additional properties into user input "
            "processed by taffy which can allow access to "
            "any data items in the DB. Taffy sets an "
            "internal index for each data item in its DB. "
            "However, it is found that the internal index "
            "can be forged by adding additional properties "
            "into user input. If index is found in the "
            "query, TaffyDB will ignore other query "
            "conditions and directly return the indexed "
            "data item. Moreover, the internal index is in "
            "an easily guessable format (e.g., "
            "T000002R000001). As such, attackers can use "
            "this vulnerability to access any data items in "
            "the DB. **Note:** `taffy` and its successor "
            "package `taffydb` are not maintained.\nNone",
            "long_description": None,
            "related_urls": [],
            "effective_severity": "HIGH",
            "source_update_time": "2023-01-30T19:22:18",
            "source_orig_time": "2020-02-19T16:43:42",
            "matched_by": "1089386|taffydb|2.6.2",
        },
        {
            "id": "CVE-2023-36665",
            "problem_type": "CWE-1321",
            "type": "npm",
            "severity": "CRITICAL",
            "cvss_score": "9.8",
            "cvss_v3": {
                "base_score": 9.8,
                "exploitability_score": 9.8,
                "impact_score": 9.8,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "REQUIRED",
                "scope": "UNCHANGED",
                "confidentiality_impact": "CRITICAL",
                "integrity_impact": "CRITICAL",
                "availability_impact": "CRITICAL",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:H/A:H",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:npm:protobufjs:*:*:*:*:*:*:*:*",
                    "package": "protobufjs",
                    "version": ">=7.0.0-<7.2.4",
                },
                "fixed_location": "7.2.4",
            },
            "short_description": "# protobufjs Prototype Pollution "
            "vulnerability\nprotobuf.js ("
            "aka protobufjs) 6.10.0 until 6.11.4 and 7.0.0 until 7.2.4 "
            "allows Prototype Pollution, a different vulnerability than "
            "CVE-2022-25878. A user-controlled protobuf message can be used "
            "by an attacker to pollute the prototype of Object.prototype by "
            "adding and overwriting its data and functions. Exploitation can "
            "involve: (1) using the function parse to parse protobuf "
            "messages on the fly, (2) loading .proto files by using "
            "load/loadSync functions, or (3) providing untrusted input to "
            "the functions ReflectionObject.setParsedOption and "
            "util.setProperty. NOTE: this CVE Record is about "
            "`Object.constructor.prototype.<new-property> = ...;` whereas "
            "CVE-2022-25878 was about `Object.__proto__.<new-property> = "
            "...;` instead.",
            "long_description": None,
            "related_urls": [
                "https://github.com/markdown-it/markdown-it/security"
                "/advisories/GHSA-6vfc-qv3f-vr6c",
                "https://nvd.nist.gov/vuln/detail/CVE-2022-21670",
                "https://github.com/markdown-it/markdown-it/commit"
                "/ffc49ab46b5b751cd2be0aabb146f2ef84986101",
                "https://github.com/markdown-it/markdown-it",
            ],
            "effective_severity": "CRITICAL",
            "source_update_time": "2023-08-15T21:16:36",
            "source_orig_time": "2023-07-05T15:30:24",
            "matched_by": "2499923747_2499958328|npm|protobufjs|7.1.2",
        },
    ]
    occs = []
    for r in res:
        vuln = CsafOccurence(r)
        occs.append(vuln)
    result = [o.to_dict() for o in occs]
    assert result == [
        {'cve': 'CVE-2019-10790',
  'cwe': {'id': 'CWE-20', 'name': 'Improper Input Validation'},
  'discovery_date': '2020-02-19T16:43:42',
  'ids': [],
  'notes': [{'audience': 'developers',
             'category': 'other',
             'text': 'Exposure of Resource to Wrong Sphere',
             'title': 'Additional CWE: CWE-668'},
            {'category': 'general',
             'details': 'Vulnerability Description',
             'text': '# TaffyDB can allow access to any data items in the DB '
                     'TaffyDB allows attackers to forge adding additional '
                     'properties into user input processed by taffy which can '
                     'allow access to any data items in the DB. Taffy sets an '
                     'internal index for each data item in its DB. However, it '
                     'is found that the internal index can be forged by adding '
                     'additional properties into user input. If index is found '
                     'in the query, TaffyDB will ignore other query conditions '
                     'and directly return the indexed data item. Moreover, the '
                     'internal index is in an easily guessable format (e.g., '
                     'T000002R000001). As such, attackers can use this '
                     'vulnerability to access any data items in the DB. '
                     '**Note:** `taffy` and its successor package `taffydb` '
                     'are not maintained. None'}],
  'product_status': {'known_affected': ['taffydb:<=2.7.3']},
  'references': []},
        {'cve': 'CVE-2023-36665',
  'cwe': {'id': 'CWE-1321',
          'name': 'Improperly Controlled Modification of Object Prototype '
                  'Attributes'},
  'discovery_date': '2023-07-05T15:30:24',
  'ids': [{'system_name': 'GitHub Advisory', 'text': 'GHSA-6vfc-qv3f-vr6c'}],
  'notes': [{'category': 'general',
             'details': 'Vulnerability Description',
             'text': '# protobufjs Prototype Pollution vulnerability '
                     'protobuf.js (aka protobufjs) 6.10.0 until 6.11.4 and '
                     '7.0.0 until 7.2.4 allows Prototype Pollution, a '
                     'different vulnerability than CVE-2022-25878. A '
                     'user-controlled protobuf message can be used by an '
                     'attacker to pollute the prototype of Object.prototype by '
                     'adding and overwriting its data and functions. '
                     'Exploitation can involve: (1) using the function parse '
                     'to parse protobuf messages on the fly, (2) loading '
                     '.proto files by using load/loadSync functions, or (3) '
                     'providing untrusted input to the functions '
                     'ReflectionObject.setParsedOption and util.setProperty. '
                     'NOTE: this CVE Record is about '
                     '`Object.constructor.prototype.<new-property> = ...;` '
                     'whereas CVE-2022-25878 was about '
                     '`Object.__proto__.<new-property> = ...;` instead.'}],
  'product_status': {'fixed': ['protobufjs:7.2.4'],
                     'known_affected': ['protobufjs:>=7.0.0-<7.2.4']},
  'references': [{'summary': 'CVE Record',
                  'url': 'https://nvd.nist.gov/vuln/detail/CVE-2022-21670'},
                 {'summary': 'GitHub Commit',
                  'url': 'https://github.com/markdown-it/markdown-it/commit/ffc49ab46b5b751cd2be0aabb146f2ef84986101'},
                 {'summary': 'GitHub Repository',
                  'url': 'https://github.com/markdown-it/markdown-it'},
                 {'summary': 'GitHub Advisory',
                  'url': 'https://github.com/markdown-it/markdown-it/security/advisories/GHSA-6vfc-qv3f-vr6c'}],
  'scores': [{'cvss_v3': {'attackVector': 'NETWORK',
                          'baseScore': 9.8,
                          'baseSeverity': 'CRITICAL',
                          'privilegesRequired': 'NONE',
                          'scope': 'UNCHANGED',
                          'userInteraction': 'REQUIRED',
                          'vectorString': 'CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:H/A:H',
                          'version': '3.1'},
              'products': ['protobufjs']}]}
    ]


def test_import_root_component():
    if os.path.exists("test/data/bom-root-comp.json"):
        [prod, ref] = import_root_component("test/data/bom-root-comp.json")
    else:
        [prod, ref] = import_root_component("data/bom-root-comp.json")

    assert prod == {
        "full_product_names": [
            {
                "name": "vuln-spring",
                "product_id": "vuln-spring:0.0.1-SNAPSHOT",
                "product_identification_helper": {
                    "purl": "pkg:maven/com.example/vuln-spring@0.0.1-SNAPSHOT"
                    "?type=jar"
                },
            }
        ]
    }

    assert ref == [
        {
            "summary": "website",
            "url": "https://projects.spring.io/spring-boot/#/spring"
            "-boot-starter-parent/vuln-spring",
        },
        {
            "summary": "vcs",
            "url": "https://github.com/spring-projects/spring-boot"
            "/spring-boot-starter-parent/vuln-spring",
        },
    ]


def test_verify_components_present():
    data = {
        "document": {
            "aggregate_severity": {"text": "Critical"},
            "category": "csaf_vex",
            "title": "Your Title",
            "csaf_version": "2.0",
            "lang": "en",
            "publisher": {
                "category": "vendor",
                "contact_details": "vendor@mcvendorson.com",
                "name": "Vendor McVendorson",
                "namespace": "https://appthreat.com",
            },
            "references": [
                {
                    "category": "self",
                    "summary": "dcksdnskljskl",
                    "url": "sdhasdjhslk.com",
                }
            ],
            "tracking": {
                "status": "draft",
                "initial_release_date": "2023-11-12T06:51:08",
                "current_release_date": "2023-11-12T06:51:08",
                "version": "1",
                "id": "2023-11-22T10:35:03_v1",
            },
        },
        "vulnerabilities": [
            {
                "id": "CVE-2020-36180",
                "problem_type": "CWE-502",
                "type": "fasterxml",
                "severity": "HIGH",
                "cvss_score": "8.1",
                "cvss_v3": {
                    "base_score": 8.1,
                    "exploitability_score": 2.2,
                    "impact_score": 5.9,
                    "attack_vector": "NETWORK",
                    "attack_complexity": "HIGH",
                    "privileges_required": "NONE",
                    "user_interaction": "NONE",
                    "scope": "UNCHANGED",
                    "confidentiality_impact": "HIGH",
                    "integrity_impact": "HIGH",
                    "availability_impact": "HIGH",
                    "vector_string": "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:U/C:H/I:H/A:H",
                },
                "package_issue": {
                    "affected_location": {
                        "cpe_uri": "cpe:2.3:a:fasterxml:jackson-databind:*:*:*:*:*:*:*:*",
                        "package": "jackson-databind",
                        "version": ">=2.7.0-<2.9.10.8",
                    },
                    "fixed_location": "2.9.10.8",
                },
                "short_description": "FasterXML jackson-databind 2.x before 2.9.10.8 mishandles the interaction between serialization gadgets and typing, related to org.apache.commons.dbcp2.cpdsadapter.DriverAdapterCPDS.",
                "long_description": None,
                "related_urls": [
                    "https://github.com/FasterXML/jackson-databind/issues/3004",
                    "https://cowtowncoder.medium.com/on-jackson-cves-dont-panic-here-is-what-you-need-to-know-54cd0d6e8062",
                ],
                "effective_severity": "HIGH",
                "source_update_time": "2023-09-13T14:56:00",
                "source_orig_time": "2021-01-07T00:15:00",
                "matched_by": "3647951461_3647986090|fasterxml|jackson-databind|2.9.6",
            },
            {
                "id": "CVE-2019-12086",
                "problem_type": "CWE-502",
                "type": "fasterxml",
                "severity": "HIGH",
                "cvss_score": "7.5",
                "cvss_v3": {
                    "base_score": 7.5,
                    "exploitability_score": 3.9,
                    "impact_score": 3.6,
                    "attack_vector": "NETWORK",
                    "attack_complexity": "LOW",
                    "privileges_required": "NONE",
                    "user_interaction": "NONE",
                    "scope": "UNCHANGED",
                    "confidentiality_impact": "HIGH",
                    "integrity_impact": "NONE",
                    "availability_impact": "NONE",
                    "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:N/A:N",
                },
                "package_issue": {
                    "affected_location": {
                        "cpe_uri": "cpe:2.3:a:fasterxml:jackson-databind:*:*:*:*:*:*:*:*",
                        "package": "jackson-databind",
                        "version": ">=2.9.0-<2.9.9",
                    },
                    "fixed_location": "2.9.9",
                },
                "short_description": "A Polymorphic Typing issue was discovered in FasterXML jackson-databind 2.x before 2.9.9. When Default Typing is enabled (either globally or for a specific property) for an externally exposed JSON endpoint, the service has the mysql-connector-java jar (8.0.14 or earlier) in the classpath, and an attacker can host a crafted MySQL server reachable by the victim, an attacker can send a crafted JSON message that allows them to read arbitrary local files on the server. This occurs because of missing com.mysql.cj.jdbc.admin.MiniAdmin validation.",
                "long_description": None,
                "related_urls": [
                    "https://www.oracle.com/security-alerts/cpujul2020.html",
                    "https://lists.apache.org/thread.html/rda99599896c3667f2cc9e9d34c7b6ef5d2bbed1f4801e1d75a2b0679@%3Ccommits.nifi.apache.org%3E",
                    "https://www.oracle.com/security-alerts/cpuoct2020.html",
                    "https://www.oracle.com/security-alerts/cpuApr2021.html",
                    "https://www.oracle.com/security-alerts/cpuapr2022.html",
                ],
                "effective_severity": "HIGH",
                "source_update_time": "2023-09-13T14:16:00",
                "source_orig_time": "2019-05-17T17:29:00",
                "matched_by": "3747044328_3747096861|fasterxml|jackson-databind|2.9.6",
            },
            {
                "id": "CVE-2018-11784",
                "problem_type": "CWE-601",
                "type": "org.apache.tomcat.embed",
                "severity": "MEDIUM",
                "cvss_score": "4.3",
                "cvss_v3": {
                    "base_score": 4.3,
                    "exploitability_score": 4.3,
                    "impact_score": 4.3,
                    "attack_vector": "NETWORK",
                    "attack_complexity": "LOW",
                    "privileges_required": "NONE",
                    "user_interaction": "REQUIRED",
                    "scope": "UNCHANGED",
                    "confidentiality_impact": "MEDIUM",
                    "integrity_impact": "MEDIUM",
                    "availability_impact": "MEDIUM",
                    "vector_string": "CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:U/C:N/I:L/A:N",
                },
                "package_issue": {
                    "affected_location": {
                        "cpe_uri": "cpe:2.3:a:org.apache.tomcat.embed:tomcat-embed-core:*:*:*:*:*:*:*:*",
                        "package": "tomcat-embed-core",
                        "version": ">=8.5.0-<8.5.34",
                    },
                    "fixed_location": "8.5.34",
                },
                "short_description": "# Moderate severity vulnerability that affects org.apache.tomcat.embed:tomcat-embed-core\nWhen the default servlet in Apache Tomcat versions 9.0.0.M1 to 9.0.11, 8.5.0 to 8.5.33 and 7.0.23 to 7.0.90 returned a redirect to a directory (e.g. redirecting to '/foo/' when the user requested '/foo') a specially crafted URL could be used to cause the redirect to be generated to any URI of the attackers choice.",
                "long_description": None,
                "related_urls": [
                    "https://nvd.nist.gov/vuln/detail/CVE-2018-11784",
                    "https://access.redhat.com/errata/RHSA-2019:0130",
                    "https://access.redhat.com/errata/RHSA-2019:0131",
                    "https://access.redhat.com/errata/RHSA-2019:0485",
                    "https://access.redhat.com/errata/RHSA-2019:1529",
                    "https://github.com/advisories/GHSA-5q99-f34m-67gc",
                    "https://seclists.org/bugtraq/2019/Dec/43",
                    "https://security.netapp.com/advisory/ntap-20181014-0002/",
                ],
                "effective_severity": "MEDIUM",
                "source_update_time": "2023-04-11T01:35:23",
                "source_orig_time": "2018-10-17T16:31:02",
                "matched_by": "2650801486_2650849443|org.apache.tomcat.embed|tomcat-embed-core|8.5.31",
            },
            {
                "id": "CVE-2022-22971",
                "problem_type": "CWE-770",
                "type": "org.springframework",
                "severity": "MEDIUM",
                "cvss_score": "6.5",
                "cvss_v3": {
                    "base_score": 6.5,
                    "exploitability_score": 6.5,
                    "impact_score": 6.5,
                    "attack_vector": "NETWORK",
                    "attack_complexity": "LOW",
                    "privileges_required": "NONE",
                    "user_interaction": "REQUIRED",
                    "scope": "UNCHANGED",
                    "confidentiality_impact": "MEDIUM",
                    "integrity_impact": "MEDIUM",
                    "availability_impact": "MEDIUM",
                    "vector_string": "CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:N/I:N/A:H",
                },
                "package_issue": {
                    "affected_location": {
                        "cpe_uri": "cpe:2.3:a:org.springframework:spring-core:*:*:*:*:*:*:*:*",
                        "package": "spring-core",
                        "version": ">=0-<5.2.22.RELEASE",
                    },
                    "fixed_location": "5.2.22.RELEASE",
                },
                "short_description": "# Allocation of Resources Without Limits or Throttling in Spring Framework\nIn spring framework versions prior to 5.3.20+ , 5.2.22+ and old unsupported versions, application with a STOMP over WebSocket endpoint is vulnerable to a denial of service attack by an authenticated user.",
                "long_description": None,
                "related_urls": [
                    "https://nvd.nist.gov/vuln/detail/CVE-2022-22971",
                    "https://security.netapp.com/advisory/ntap-20220616-0003/",
                    "https://tanzu.vmware.com/security/cve-2022-22971",
                    "https://www.oracle.com/security-alerts/cpujul2022.html",
                ],
                "effective_severity": "MEDIUM",
                "source_update_time": "2023-04-11T01:33:53",
                "source_orig_time": "2022-05-13T00:00:29",
                "matched_by": "2660437234_2660469022|org.springframework|spring-core|5.0.7.RELEASE",
            },
            {
                "id": "CVE-2022-40150",
                "problem_type": "CWE-400,CWE-674",
                "type": "org.codehaus.jettison",
                "severity": "HIGH",
                "cvss_score": "7.5",
                "cvss_v3": {
                    "base_score": 7.5,
                    "exploitability_score": 7.5,
                    "impact_score": 7.5,
                    "attack_vector": "NETWORK",
                    "attack_complexity": "LOW",
                    "privileges_required": "NONE",
                    "user_interaction": "REQUIRED",
                    "scope": "UNCHANGED",
                    "confidentiality_impact": "HIGH",
                    "integrity_impact": "HIGH",
                    "availability_impact": "HIGH",
                    "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H",
                },
                "package_issue": {
                    "affected_location": {
                        "cpe_uri": "cpe:2.3:a:org.codehaus.jettison:jettison:*:*:*:*:*:*:*:*",
                        "package": "jettison",
                        "version": ">=0-<1.5.2",
                    },
                    "fixed_location": "1.5.2",
                },
                "short_description": "# Jettison memory exhaustion\nThose using Jettison to parse untrusted XML or JSON data may be vulnerable to Denial of Service attacks (DOS). If the parser is running on user supplied input, an attacker may supply content that causes the parser to crash by Out of memory. This effect may support a denial of service attack.",
                "long_description": None,
                "related_urls": [
                    "https://nvd.nist.gov/vuln/detail/CVE-2022-40150",
                    "https://github.com/jettison-json/jettison/issues/45",
                    "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=46549",
                    "https://github.com/jettison-json/jettison",
                    "https://lists.debian.org/debian-lts-announce/2022/12/msg00045.html",
                    "https://www.debian.org/security/2023/dsa-5312",
                ],
                "effective_severity": "HIGH",
                "source_update_time": "2023-07-13T19:19:12",
                "source_orig_time": "2022-09-17T00:00:41",
                "matched_by": "2662034264_2662079740|org.codehaus.jettison|jettison|1.3.7",
            },
        ],
    }
    metadata = {
        "depscan_version": "4.3.3",
        "note": [
            {"audience": "", "category": "", "text": "", "title": ""},
            {"audience": "", "category": "", "text": "", "title": ""},
        ],
        "reference": [
            {
                "category": "self",
                "summary": "dcksdnskljskl",
                "url": "sdhasdjhslk.com",
            },
            {"category": "", "summary": "", "url": ""},
        ],
        "distribution": {"label": "", "text": "", "url": ""},
        "document": {"category": "csaf_vex", "title": "Your Title"},
        "product_tree": {"easy_import": ""},
        "publisher": {
            "category": "vendor",
            "contact_details": "vendor@mcvendorson.com",
            "name": "Vendor McVendorson",
            "namespace": "https://appthreat.com",
        },
        "tracking": {
            "status": "draft",
            "initial_release_date": "2023-11-12T06:51:08",
            "current_release_date": "2023-11-12T06:51:08",
            "version": "1",
            "id": "",
        },
    }
    if os.path.exists("test/data/bom-root-comp.json"):
        vdr_file = "test/data/bom-root-comp.json"
    else:
        vdr_file = "data/bom-root-comp.json"

    [template, new_metadata] = verify_components_present(
        data, metadata, vdr_file
    )
    assert template["document"]["notes"] == [
        {
            "category": "legal_disclaimer",
            "text": "Depscan reachable code only covers the "
            "project source code, not the code of "
            "dependencies. A dependency may execute "
            "vulnerable code when called even if it is "
            "not in the project's source code. Regard the "
            "Depscan-set flag of "
            "'code_not_in_execute_path' with this in "
            "mind.",
        }
    ]
    assert template["product_tree"] == {
        "full_product_names": [
            {
                "name": "vuln-spring",
                "product_id": "vuln-spring:0.0.1-SNAPSHOT",
                "product_identification_helper": {
                    "purl": "pkg:maven/com.example/vuln-spring@0.0.1-SNAPSHOT?type=jar"
                },
            }
        ]
    }

    assert new_metadata["tracking"] == {
        "current_release_date": "2023-11-12T06:51:08",
        "id": "",
        "initial_release_date": "2023-11-12T06:51:08",
        "status": "draft",
        "version": "1",
    }


def test_add_vulnerabilities():
    data = {
        "document": {
            "aggregate_severity": {},
            "category": "csaf_vex",
            "title": "Your Title",
            "csaf_version": "2.0",
            "distribution": {"label": "", "text": "", "url": ""},
            "lang": "en",
            "notes": [
                {"audience": "", "category": "", "text": "", "title": ""},
                {"audience": "", "category": "", "text": "", "title": ""},
            ],
            "publisher": {
                "category": "vendor",
                "contact_details": "vendor@mcvendorson.com",
                "name": "Vendor McVendorson",
                "namespace": "https://appthreat.com",
            },
            "references": [
                {
                    "category": "self",
                    "summary": "dcksdnskljskl",
                    "url": "sdhasdjhslk.com",
                },
                {"category": "", "summary": "", "url": ""},
            ],
            "tracking": {
                "status": "draft",
                "initial_release_date": "2023-11-12T06:51:08",
                "current_release_date": "2023-11-12T06:51:08",
                "version": "1",
                "id": "2023-11-21T21:39:14_v1",
                "revision_history": [],
            },
        },
        "product_tree": None,
        "vulnerabilities": [],
    }
    reached_purls = {
        "pkg:maven/org.apache.tomcat.embed/tomcat-embed-core@8.5.31"
        "?type=jar": 3,
        "pkg:maven/org.codehaus.jettison/jettison@1.3.7?type=jar": 19,
    }
    direct_purls = {}
    results = [
        {
            "id": "CVE-2020-36180",
            "problem_type": "CWE-502",
            "type": "fasterxml",
            "severity": "UNKNOWN",
            "cvss_score": "8.1",
            "cvss_v3": {
                "base_score": 8.1,
                "exploitability_score": 2.2,
                "impact_score": 5.9,
                "attack_vector": "NETWORK",
                "attack_complexity": "HIGH",
                "privileges_required": "NONE",
                "user_interaction": "NONE",
                "scope": "UNCHANGED",
                "confidentiality_impact": "HIGH",
                "integrity_impact": "HIGH",
                "availability_impact": "HIGH",
                "vector_string": "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:U/C:H/I:H/A:H",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:fasterxml:jackson-databind:*:*:*:*:*:*:*:*",
                    "package": "jackson-databind",
                    "version": ">=2.7.0-<2.9.10.8",
                },
                "fixed_location": "2.9.10.8",
            },
            "short_description": "FasterXML jackson-databind 2.x before 2.9.10.8 mishandles the interaction between serialization gadgets and typing, related to org.apache.commons.dbcp2.cpdsadapter.DriverAdapterCPDS.",
            "long_description": None,
            "related_urls": [
                "https://github.com/FasterXML/jackson-databind/issues/3004",
                "https://cowtowncoder.medium.com/on-jackson-cves-dont-panic-here-is-what-you-need-to-know-54cd0d6e8062",
            ],
            "effective_severity": "HIGH",
            "source_update_time": "2023-09-13T14:56:00",
            "source_orig_time": "2021-01-07T00:15:00",
            "matched_by": "3647951461_3647986090|fasterxml|jackson-databind|2.9.6",
        },
        {
            "id": "CVE-2019-12086",
            "problem_type": "CWE-502",
            "type": "fasterxml",
            "severity": "LOW",
            "cvss_score": "7.5",
            "cvss_v3": {
                "base_score": 7.5,
                "exploitability_score": 3.9,
                "impact_score": 3.6,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "NONE",
                "scope": "UNCHANGED",
                "confidentiality_impact": "HIGH",
                "integrity_impact": "NONE",
                "availability_impact": "NONE",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:N/A:N",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:fasterxml:jackson-databind:*:*:*:*:*:*:*:*",
                    "package": "jackson-databind",
                    "version": ">=2.9.0-<2.9.9",
                },
                "fixed_location": "2.9.9",
            },
            "short_description": "A Polymorphic Typing issue was discovered in FasterXML jackson-databind 2.x before 2.9.9. When Default Typing is enabled (either globally or for a specific property) for an externally exposed JSON endpoint, the service has the mysql-connector-java jar (8.0.14 or earlier) in the classpath, and an attacker can host a crafted MySQL server reachable by the victim, an attacker can send a crafted JSON message that allows them to read arbitrary local files on the server. This occurs because of missing com.mysql.cj.jdbc.admin.MiniAdmin validation.",
            "long_description": None,
            "related_urls": [
                "https://www.oracle.com/security-alerts/cpujul2020.html",
                "https://lists.apache.org/thread.html/rda99599896c3667f2cc9e9d34c7b6ef5d2bbed1f4801e1d75a2b0679@%3Ccommits.nifi.apache.org%3E",
                "https://www.oracle.com/security-alerts/cpuoct2020.html",
                "https://www.oracle.com/security-alerts/cpuApr2021.html",
                "https://www.oracle.com/security-alerts/cpuapr2022.html",
            ],
            "effective_severity": "HIGH",
            "source_update_time": "2023-09-13T14:16:00",
            "source_orig_time": "2019-05-17T17:29:00",
            "matched_by": "3747044328_3747096861|fasterxml|jackson-databind|2.9.6",
        },
        {
            "id": "CVE-2018-11784",
            "problem_type": "CWE-601",
            "type": "org.apache.tomcat.embed",
            "severity": "MEDIUM",
            "cvss_score": "4.3",
            "cvss_v3": {
                "base_score": 4.3,
                "exploitability_score": 4.3,
                "impact_score": 4.3,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "REQUIRED",
                "scope": "UNCHANGED",
                "confidentiality_impact": "MEDIUM",
                "integrity_impact": "MEDIUM",
                "availability_impact": "MEDIUM",
                "vector_string": "CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:U/C:N/I:L/A:N",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:org.apache.tomcat.embed:tomcat-embed-core:*:*:*:*:*:*:*:*",
                    "package": "tomcat-embed-core",
                    "version": ">=8.5.0-<8.5.34",
                },
                "fixed_location": "8.5.34",
            },
            "short_description": "# Moderate severity vulnerability that affects org.apache.tomcat.embed:tomcat-embed-core\nWhen the default servlet in Apache Tomcat versions 9.0.0.M1 to 9.0.11, 8.5.0 to 8.5.33 and 7.0.23 to 7.0.90 returned a redirect to a directory (e.g. redirecting to '/foo/' when the user requested '/foo') a specially crafted URL could be used to cause the redirect to be generated to any URI of the attackers choice.",
            "long_description": None,
            "related_urls": [
                "https://nvd.nist.gov/vuln/detail/CVE-2018-11784",
                "https://access.redhat.com/errata/RHSA-2019:0130",
                "https://access.redhat.com/errata/RHSA-2019:0131",
                "https://access.redhat.com/errata/RHSA-2019:0485",
                "https://access.redhat.com/errata/RHSA-2019:1529",
                "https://github.com/advisories/GHSA-5q99-f34m-67gc",
                "https://seclists.org/bugtraq/2019/Dec/43",
                "https://security.netapp.com/advisory/ntap-20181014-0002/",
            ],
            "effective_severity": "MEDIUM",
            "source_update_time": "2023-04-11T01:35:23",
            "source_orig_time": "2018-10-17T16:31:02",
            "matched_by": "2650801486_2650849443|org.apache.tomcat.embed|tomcat-embed-core|8.5.31",
        },
        {
            "id": "CVE-2022-22971",
            "problem_type": "CWE-770",
            "type": "org.springframework",
            "severity": "MEDIUM",
            "cvss_score": "6.5",
            "cvss_v3": {
                "base_score": 6.5,
                "exploitability_score": 6.5,
                "impact_score": 6.5,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "REQUIRED",
                "scope": "UNCHANGED",
                "confidentiality_impact": "MEDIUM",
                "integrity_impact": "MEDIUM",
                "availability_impact": "MEDIUM",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:N/I:N/A:H",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:org.springframework:spring-core:*:*:*:*:*:*:*:*",
                    "package": "spring-core",
                    "version": ">=0-<5.2.22.RELEASE",
                },
                "fixed_location": "5.2.22.RELEASE",
            },
            "short_description": "# Allocation of Resources Without Limits or Throttling in Spring Framework\nIn spring framework versions prior to 5.3.20+ , 5.2.22+ and old unsupported versions, application with a STOMP over WebSocket endpoint is vulnerable to a denial of service attack by an authenticated user.",
            "long_description": None,
            "related_urls": [
                "https://nvd.nist.gov/vuln/detail/CVE-2022-22971",
                "https://security.netapp.com/advisory/ntap-20220616-0003/",
                "https://tanzu.vmware.com/security/cve-2022-22971",
                "https://www.oracle.com/security-alerts/cpujul2022.html",
            ],
            "effective_severity": "MEDIUM",
            "source_update_time": "2023-04-11T01:33:53",
            "source_orig_time": "2022-05-13T00:00:29",
            "matched_by": "2660437234_2660469022|org.springframework|spring-core|5.0.7.RELEASE",
        },
        {
            "id": "CVE-2022-40150",
            "problem_type": "CWE-400,CWE-674",
            "type": "org.codehaus.jettison",
            "severity": "LOW",
            "cvss_score": "7.5",
            "cvss_v3": {
                "base_score": 7.5,
                "exploitability_score": 7.5,
                "impact_score": 7.5,
                "attack_vector": "NETWORK",
                "attack_complexity": "LOW",
                "privileges_required": "NONE",
                "user_interaction": "REQUIRED",
                "scope": "UNCHANGED",
                "confidentiality_impact": "HIGH",
                "integrity_impact": "HIGH",
                "availability_impact": "HIGH",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H",
            },
            "package_issue": {
                "affected_location": {
                    "cpe_uri": "cpe:2.3:a:org.codehaus.jettison:jettison:*:*:*:*:*:*:*:*",
                    "package": "jettison",
                    "version": ">=0-<1.5.2",
                },
                "fixed_location": "1.5.2",
            },
            "short_description": "# Jettison memory exhaustion\nThose using Jettison to parse untrusted XML or JSON data may be vulnerable to Denial of Service attacks (DOS). If the parser is running on user supplied input, an attacker may supply content that causes the parser to crash by Out of memory. This effect may support a denial of service attack.",
            "long_description": None,
            "related_urls": [
                "https://nvd.nist.gov/vuln/detail/CVE-2022-40150",
                "https://github.com/jettison-json/jettison/issues/45",
                "https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=46549",
                "https://github.com/jettison-json/jettison",
                "https://lists.debian.org/debian-lts-announce/2022/12/msg00045.html",
                "https://www.debian.org/security/2023/dsa-5312",
            ],
            "effective_severity": "HIGH",
            "source_update_time": "2023-07-13T19:19:12",
            "source_orig_time": "2022-09-17T00:00:41",
            "matched_by": "2662034264_2662079740|org.codehaus.jettison|jettison|1.3.7",
        },
    ]
    new_results = add_vulnerabilities(
        data, results, direct_purls, reached_purls
    )

    assert new_results["document"]["aggregate_severity"].get("text") == "Medium"

    assert new_results.get("vulnerabilities") == [{
        'cve': 'CVE-2020-36180',
        'cwe': {'id': 'CWE-502', 'name': 'Deserialization of Untrusted Data'},
        'discovery_date': '2021-01-07T00:15:00',
        'flags': [{'label': 'vulnerable_code_not_in_execute_path'}],
        'ids': [{'system_name': 'GitHub Issue [FasterXML/jackson-databind]',
                 'text': '3004'}],
        'notes': [{'category': 'general',
                   'details': 'Vulnerability Description',
                   'text': 'FasterXML jackson-databind 2.x before 2.9.10.8 '
                           'mishandles the interaction between serialization gadgets '
                           'and typing, related to '
                           'org.apache.commons.dbcp2.cpdsadapter.DriverAdapterCPDS.'}],
        'product_status': {'fixed': ['jackson-databind:2.9.10.8'],
                           'known_affected': ['jackson-databind:>=2.7.0-<2.9.10.8']},
        'references': [{'summary': 'Other',
                        'url': 'https://cowtowncoder.medium.com/on-jackson-cves-dont-panic-here-is-what-you-need-to-know-54cd0d6e8062'},
                       {'summary': 'GitHub Issue',
                        'url': 'https://github.com/FasterXML/jackson-databind/issues/3004'}]},
        {'cve': 'CVE-2019-12086',
         'cwe': {'id': 'CWE-502', 'name': 'Deserialization of Untrusted Data'},
         'discovery_date': '2019-05-17T17:29:00',
         'flags': [{'label': 'vulnerable_code_not_in_execute_path'}],
         'ids': [],
         'notes': [{'category': 'general',
                    'details': 'Vulnerability Description',
                    'text': 'A Polymorphic Typing issue was discovered in FasterXML '
                            'jackson-databind 2.x before 2.9.9. When Default Typing '
                            'is enabled (either globally or for a specific property) '
                            'for an externally exposed JSON endpoint, the service has '
                            'the mysql-connector-java jar (8.0.14 or earlier) in the '
                            'classpath, and an attacker can host a crafted MySQL '
                            'server reachable by the victim, an attacker can send a '
                            'crafted JSON message that allows them to read arbitrary '
                            'local files on the server. This occurs because of '
                            'missing com.mysql.cj.jdbc.admin.MiniAdmin validation.'}],
         'product_status': {'fixed': ['jackson-databind:2.9.9'],
                            'known_affected': ['jackson-databind:>=2.9.0-<2.9.9']},
         'references': [{'summary': 'Oracle Security Alert',
                         'url': 'https://www.oracle.com/security-alerts/cpujul2020.html'},
                        {'summary': 'Mailing List Other',
                         'url': 'https://lists.apache.org/thread.html/rda99599896c3667f2cc9e9d34c7b6ef5d2bbed1f4801e1d75a2b0679@%3Ccommits.nifi.apache.org%3E'},
                        {'summary': 'Oracle Security Alert',
                         'url': 'https://www.oracle.com/security-alerts/cpuoct2020.html'},
                        {'summary': 'Oracle Security Alert',
                         'url': 'https://www.oracle.com/security-alerts/cpuApr2021.html'},
                        {'summary': 'Oracle Security Alert',
                         'url': 'https://www.oracle.com/security-alerts/cpuapr2022.html'}],
         'scores': [{'cvss_v3': {'attackVector': 'NETWORK',
                                 'baseScore': 7.5,
                                 'baseSeverity': 'LOW',
                                 'privilegesRequired': 'NONE',
                                 'scope': 'UNCHANGED',
                                 'userInteraction': 'NONE',
                                 'vectorString': 'CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:H/I:N/A:N',
                                 'version': '3.1'},
                     'products': ['jackson-databind']}]},
        {'cve': 'CVE-2018-11784',
         'cwe': {'id': 'CWE-601', 'name': 'URL Redirection to Untrusted Site'},
         'discovery_date': '2018-10-17T16:31:02',
         'ids': [{'system_name': 'GitHub Advisory', 'text': 'GHSA-5q99-f34m-67gc'},
                 {'system_name': 'Red Hat Advisory', 'text': 'RHSA-2019:0130'},
                 {'system_name': 'Red Hat Advisory', 'text': 'RHSA-2019:0131'},
                 {'system_name': 'Red Hat Advisory', 'text': 'RHSA-2019:0485'},
                 {'system_name': 'Red Hat Advisory', 'text': 'RHSA-2019:1529'},
                 {'system_name': 'NetApp Advisory', 'text': 'ntap-20181014-0002'}],
         'notes': [{'category': 'general',
                    'details': 'Vulnerability Description',
                    'text': '# Moderate severity vulnerability that affects '
                            'org.apache.tomcat.embed:tomcat-embed-core When the '
                            'default servlet in Apache Tomcat versions 9.0.0.M1 to '
                            '9.0.11, 8.5.0 to 8.5.33 and 7.0.23 to 7.0.90 returned a '
                            "redirect to a directory (e.g. redirecting to '/foo/' "
                            "when the user requested '/foo') a specially crafted URL "
                            'could be used to cause the redirect to be generated to '
                            'any URI of the attackers choice.'}],
         'product_status': {'fixed': ['tomcat-embed-core:8.5.34'],
                            'known_affected': ['tomcat-embed-core:>=8.5.0-<8.5.34']},
         'references': [{'summary': 'CVE Record',
                         'url': 'https://nvd.nist.gov/vuln/detail/CVE-2018-11784'},
                        {'summary': 'Other',
                         'url': 'https://seclists.org/bugtraq/2019/Dec/43'},
                        {'summary': 'Red Hat Advisory',
                         'url': 'https://access.redhat.com/errata/RHSA-2019:0130'},
                        {'summary': 'Red Hat Advisory',
                         'url': 'https://access.redhat.com/errata/RHSA-2019:0131'},
                        {'summary': 'Red Hat Advisory',
                         'url': 'https://access.redhat.com/errata/RHSA-2019:0485'},
                        {'summary': 'Red Hat Advisory',
                         'url': 'https://access.redhat.com/errata/RHSA-2019:1529'},
                        {'summary': 'GitHub Advisory',
                         'url': 'https://github.com/advisories/GHSA-5q99-f34m-67gc'},
                        {'summary': 'NetApp Advisory',
                         'url': 'https://security.netapp.com/advisory/ntap-20181014-0002/'}],
         'scores': [{'cvss_v3': {'attackVector': 'NETWORK',
                                 'baseScore': 4.3,
                                 'baseSeverity': 'MEDIUM',
                                 'privilegesRequired': 'NONE',
                                 'scope': 'UNCHANGED',
                                 'userInteraction': 'REQUIRED',
                                 'vectorString': 'CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:U/C:N/I:L/A:N',
                                 'version': '3.0'},
                     'products': ['tomcat-embed-core']}]},
        {'cve': 'CVE-2022-22971',
         'cwe': {'id': 'CWE-770',
                 'name': 'Allocation of Resources Without Limits or Throttling'},
         'discovery_date': '2022-05-13T00:00:29',
         'flags': [{'label': 'vulnerable_code_not_in_execute_path'}],
         'ids': [{'system_name': 'NetApp Advisory', 'text': 'ntap-20220616-0003'}],
         'notes': [{'category': 'general',
                    'details': 'Vulnerability Description',
                    'text': '# Allocation of Resources Without Limits or Throttling '
                            'in Spring Framework In spring framework versions prior '
                            'to 5.3.20+ , 5.2.22+ and old unsupported versions, '
                            'application with a STOMP over WebSocket endpoint is '
                            'vulnerable to a denial of service attack by an '
                            'authenticated user.'}],
         'product_status': {'fixed': ['spring-core:5.2.22.RELEASE'],
                            'known_affected': ['spring-core:>=0-<5.2.22.RELEASE']},
         'references': [{'summary': 'CVE Record',
                         'url': 'https://nvd.nist.gov/vuln/detail/CVE-2022-22971'},
                        {'summary': 'CVE Record',
                         'url': 'https://tanzu.vmware.com/security/cve-2022-22971'},
                        {'summary': 'Oracle Security Alert',
                         'url': 'https://www.oracle.com/security-alerts/cpujul2022.html'},
                        {'summary': 'NetApp Advisory',
                         'url': 'https://security.netapp.com/advisory/ntap-20220616-0003/'}],
         'scores': [{'cvss_v3': {'attackVector': 'NETWORK',
                                 'baseScore': 6.5,
                                 'baseSeverity': 'MEDIUM',
                                 'privilegesRequired': 'NONE',
                                 'scope': 'UNCHANGED',
                                 'userInteraction': 'REQUIRED',
                                 'vectorString': 'CVSS:3.1/AV:N/AC:L/PR:L/UI:N/S:U/C:N/I:N/A:H',
                                 'version': '3.1'},
                     'products': ['spring-core']}]},
        {'cve': 'CVE-2022-40150',
         'cwe': {'id': 'CWE-400', 'name': 'Uncontrolled Resource Consumption'},
         'discovery_date': '2022-09-17T00:00:41',
         'ids': [{'system_name': 'GitHub Issue [jettison-json/jettison]',
                  'text': '45'},
                 {'system_name': 'Chromium Issue [oss-fuzz]', 'text': '46549'},
                 {'system_name': 'Debian Advisory', 'text': 'dsa-5312'}],
         'notes': [{'audience': 'developers',
                    'category': 'other',
                    'text': 'Uncontrolled Recursion',
                    'title': 'Additional CWE: CWE-674'},
                   {'category': 'general',
                    'details': 'Vulnerability Description',
                    'text': '# Jettison memory exhaustion Those using Jettison to '
                            'parse untrusted XML or JSON data may be vulnerable to '
                            'Denial of Service attacks (DOS). If the parser is '
                            'running on user supplied input, an attacker may supply '
                            'content that causes the parser to crash by Out of '
                            'memory. This effect may support a denial of service '
                            'attack.'}],
         'product_status': {'fixed': ['jettison:1.5.2'],
                            'known_affected': ['jettison:>=0-<1.5.2']},
         'references': [{'summary': 'CVE Record',
                         'url': 'https://nvd.nist.gov/vuln/detail/CVE-2022-40150'},
                        {'summary': 'GitHub Repository',
                         'url': 'https://github.com/jettison-json/jettison'},
                        {'summary': 'Mailing List Announcement',
                         'url': 'https://lists.debian.org/debian-lts-announce/2022/12/msg00045.html'},
                        {'summary': 'GitHub Issue',
                         'url': 'https://github.com/jettison-json/jettison/issues/45'},
                        {'summary': 'Chromium Issue',
                         'url': 'https://bugs.chromium.org/p/oss-fuzz/issues/detail?id=46549'},
                        {'summary': 'Debian Advisory',
                         'url': 'https://www.debian.org/security/2023/dsa-5312'}],
         'scores': [{'cvss_v3': {'attackVector': 'NETWORK',
                                 'baseScore': 7.5,
                                 'baseSeverity': 'LOW',
                                 'privilegesRequired': 'NONE',
                                 'scope': 'UNCHANGED',
                                 'userInteraction': 'REQUIRED',
                                 'vectorString': 'CVSS:3.1/AV:N/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:H',
                                 'version': '3.1'},
                     'products': ['jettison']}]}
                                                  ]


def test_version_helper():
    # Returns True if the version in 'reached' satisfies the conditions
    # specified by 'vdata'.
    reached = ["1.0.0", "2.0.0", "3.0.0"]
    vdata = {"lower": "2.0.0", "lmod": ">=", "upper": "3.0.0", "umod": "<"}
    result = version_helper(reached, vdata)
    assert result is True

    # Returns False if the version in 'reached' does not satisfy the
    # conditions specified by 'vdata'.
    vdata = {"lower": "4.0.0", "lmod": ">=", "upper": "5.0.0", "umod": "<"}
    result = version_helper(reached, vdata)
    assert result is False

    # Returns True if the lower bound modifier is ">=" and the version in
    # 'reached' is equal to the lower bound of the version range.
    vdata = {"lower": "2.0.0", "lmod": ">=", "upper": "3.0.0", "umod": "<"}
    result = version_helper(reached, vdata)
    assert result is True

    # Returns False if the lower bound modifier is ">=" and the version in
    # 'reached' is less than the lower bound of the version range.
    vdata = {"lower": "4.0.0", "lmod": ">=", "upper": "5.0.0", "umod": "<"}
    result = version_helper(reached, vdata)
    assert result is False

    # Returns False if the upper bound modifier is "<=" and the version in
    # 'reached' is greater than the upper bound of the version range.
    vdata = {"lower": "2.0.0", "lmod": ">=", "upper": "3.0.0", "umod": "<="}
    result = version_helper(reached, vdata)
    assert result is True

    #  Returns False if the lower bound modifier is ">" and the version in
    # 'reached' is less than or equal to the lower bound of the version range.
    vdata = {"lower": "2.0.0", "lmod": ">", "upper": "3.0.0", "umod": "<"}
    result = version_helper(reached, vdata)
    assert result is False



import math
from datetime import datetime

import httpx
from rich.progress import Progress

from depscan.lib import config
from depscan.lib.logger import LOG, console


def get_lookup_url(registry_type, pkg):
    """
    Generating the lookup URL based on the registry type and package
    information.

    :param registry_type: The type of registry ("npm" or "pypi")
    :param pkg: Dict or string of package information
    :returns: Package name, lookup URL
    """
    vendor = None
    if isinstance(pkg, dict):
        vendor = pkg.get("vendor")
        name = pkg.get("name")
    else:
        tmpA = pkg.split("|")
        name = tmpA[len(tmpA) - 2]
        if len(tmpA) == 3:
            vendor = tmpA[0]
    key = name
    # Prefix vendor for npm
    if registry_type == "npm":
        if vendor and vendor != "npm":
            # npm expects namespaces to start with an @
            if not vendor.startswith("@"):
                vendor = "@" + vendor
            key = f"{vendor}/{name}"
        return key, f"{config.npm_server}/{key}"
    elif registry_type == "pypi":
        return key, f"{config.pypi_server}/{key}/json"
    return None, None


def metadata_from_registry(registry_type, scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query registry for the package metadata

    :param registry_type: The type of registry to query
    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return:  A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    metadata_dict = {}
    # Circuit breaker flag to break the risk audit in case of many api errors
    circuit_breaker = False
    # Track the api failures count
    failure_count = 0
    done_count = 0
    with Progress(
        console=console,
        transient=True,
        redirect_stderr=False,
        redirect_stdout=False,
        refresh_per_second=1,
    ) as progress:
        task = progress.add_task("[green] Auditing packages", total=len(pkg_list))
        for pkg in pkg_list:
            if circuit_breaker:
                LOG.info(
                    "Risk audited has been interrupted due to frequent api "
                    "errors. Please try again later."
                )
                progress.stop()
                return {}
            scope = pkg.get("scope", "").lower()
            key, lookup_url = get_lookup_url(registry_type, pkg)
            if not key or not lookup_url or key.startswith("https://"):
                progress.advance(task)
                continue
            progress.update(task, description=f"Checking {key}")
            try:
                r = httpx.get(
                    url=lookup_url,
                    follow_redirects=True,
                    timeout=config.request_timeout_sec,
                )
                json_data = r.json()
                # Npm returns this error if the package is not found
                if (
                    json_data.get("code") == "MethodNotAllowedError"
                    or r.status_code > 400
                ):
                    continue
                is_private_pkg = False
                if private_ns:
                    namespace_prefixes = private_ns.split(",")
                    for ns in namespace_prefixes:
                        if key.lower().startswith(ns.lower()) or key.lower().startswith(
                            "@" + ns.lower()
                        ):
                            is_private_pkg = True
                            break
                risk_metrics = {}
                if registry_type == "npm":
                    risk_metrics = npm_pkg_risk(json_data, is_private_pkg, scope)
                elif registry_type == "pypi":
                    project_type_pkg = f"python:{key}".lower()
                    required_pkgs = scoped_pkgs.get("required", [])
                    optional_pkgs = scoped_pkgs.get("optional", [])
                    excluded_pkgs = scoped_pkgs.get("excluded", [])
                    if (
                        pkg.get("purl") in required_pkgs
                        or project_type_pkg in required_pkgs
                    ):
                        scope = "required"
                    elif (
                        pkg.get("purl") in optional_pkgs
                        or project_type_pkg in optional_pkgs
                    ):
                        scope = "optional"
                    elif (
                        pkg.get("purl") in excluded_pkgs
                        or project_type_pkg in excluded_pkgs
                    ):
                        scope = "excluded"
                    risk_metrics = pypi_pkg_risk(json_data, is_private_pkg, scope)
                metadata_dict[key] = {
                    "scope": scope,
                    "pkg_metadata": json_data,
                    "risk_metrics": risk_metrics,
                    "is_private_pkg": is_private_pkg,
                }
            except Exception as e:
                LOG.debug(e)
                failure_count += 1
            progress.advance(task)
            done_count += 1
            if failure_count >= config.max_request_failures:
                circuit_breaker = True
    LOG.debug(
        "Retrieved package metadata for %d/%d packages. Failures count %d",
        done_count,
        len(pkg_list),
        failure_count,
    )
    return metadata_dict


def npm_metadata(scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query npm for the package metadata

    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return: A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    return metadata_from_registry("npm", scoped_pkgs, pkg_list, private_ns)


def pypi_metadata(scoped_pkgs, pkg_list, private_ns=None):
    """
    Method to query pypi for the package metadata

    :param scoped_pkgs: Dictionary of lists of packages per scope
    :param pkg_list: List of package dictionaries
    :param private_ns: Private namespace
    :return: A dict of package metadata, risk metrics, and private package
    flag for each package
    """
    return metadata_from_registry("pypi", scoped_pkgs, pkg_list, private_ns)


def get_category_score(
    param, max_value=config.default_max_value, weight=config.default_weight
):
    """
    Return parameter score given its current value, max value and
    parameter weight.

    :param param: The current value of the parameter
    :param max_value: The maximum value of the parameter
    :param weight: The weight of the parameter
    :return: The calculated score as a float value
    """
    try:
        param = float(param)
    except ValueError:
        param = 0
    try:
        max_value = float(max_value)
    except ValueError:
        max_value = config.default_max_value
    try:
        weight = float(weight)
    except ValueError:
        weight = config.default_weight
    return (math.log(1 + param) / math.log(1 + max(param, max_value))) * weight


def calculate_risk_score(risk_metrics):
    """
    Method to calculate a total risk score based on risk metrics. This is
    based on a weighted formula and might require customization based on use
    cases

    :param risk_metrics: Dict containing many risk metrics
    :return: The calculated total risk score
    """
    if not risk_metrics:
        return 0
    num_risks = 0
    working_score = 0
    total_weight = 0
    for k, v in risk_metrics.items():
        # Is the _risk key set to True
        if k.endswith("_risk") and v is True:
            risk_category = k.replace("_risk", "")
            risk_category_value = risk_metrics.get(f"{risk_category}_value", 0)
            risk_category_max = getattr(
                config, f"{risk_category}_max", config.default_max_value
            )
            risk_category_weight = getattr(
                config, f"{risk_category}_weight", config.default_weight
            )
            risk_category_base = getattr(config, f"{risk_category}", 0)
            value = risk_category_value
            if (
                risk_category_base
                and (
                    isinstance(risk_category_base, float)
                    or isinstance(risk_category_base, int)
                )
                and risk_category_base > risk_category_value
            ):
                value = risk_category_base - risk_category_value
            elif risk_category_max and risk_category_max > risk_category_value:
                value = risk_category_max - risk_category_value
            cat_score = get_category_score(
                value, risk_category_max, risk_category_weight
            )
            total_weight += risk_category_weight
            working_score += min(cat_score, 1)
            num_risks += 1
    working_score = round(working_score * total_weight / config.total_weight, 5)
    working_score = max(min(working_score, 1), 0)
    return working_score


def compute_time_risks(
    risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
):
    """
    Compute risks based on creation, modified and time elapsed

    :param risk_metrics: A dict containing the risk metrics for the package.
    :param created_now_diff: Time difference from creation of the package and
    the current time.
    :param mod_create_diff: Time difference from
    modification and creation of the package.
    :param latest_now_diff: Time difference between the latest version of the
    package and the current
    time.
    :return: The updated risk metrics dictionary with the calculated
    risks and values.
    """
    # Check if the package is at least 1 year old. Quarantine period.
    if created_now_diff.total_seconds() < config.created_now_quarantine_seconds:
        risk_metrics["created_now_quarantine_seconds_risk"] = True
        risk_metrics[
            "created_now_quarantine_seconds_value"
        ] = latest_now_diff.total_seconds()

    # Check for the maximum seconds difference between latest version and now
    if latest_now_diff.total_seconds() > config.latest_now_max_seconds:
        risk_metrics["latest_now_max_seconds_risk"] = True
        risk_metrics["latest_now_max_seconds_value"] = latest_now_diff.total_seconds()
        # Since the package is quite old we can relax the min versions risk
        risk_metrics["pkg_min_versions_risk"] = False
    else:
        # Check for the minimum seconds difference between creation and
        # modified date This check catches several old npm packages that was
        # created and immediately updated within a day To reduce noise we
        # check for the age first and perform this check only for newish
        # packages
        if mod_create_diff.total_seconds() < config.mod_create_min_seconds:
            risk_metrics["mod_create_min_seconds_risk"] = True
            risk_metrics[
                "mod_create_min_seconds_value"
            ] = mod_create_diff.total_seconds()
    # Check for the minimum seconds difference between latest version and now
    if latest_now_diff.total_seconds() < config.latest_now_min_seconds:
        risk_metrics["latest_now_min_seconds_risk"] = True
        risk_metrics["latest_now_min_seconds_value"] = latest_now_diff.total_seconds()
    return risk_metrics


def pypi_pkg_risk(pkg_metadata, is_private_pkg, scope):
    """
    Calculate various package risks based on the metadata from pypi.

    :param pkg_metadata: A dict containing the metadata of the package from PyPI
    :param is_private_pkg: Boolean to indicate if this package is private
    :param scope: Package scope
    :return: Dict of risk metrics and corresponding PyPI values.
    """
    risk_metrics = {
        "pkg_deprecated_risk": False,
        "pkg_min_versions_risk": False,
        "created_now_quarantine_seconds_risk": False,
        "latest_now_max_seconds_risk": False,
        "mod_create_min_seconds_risk": False,
        "pkg_min_maintainers_risk": False,
        "pkg_private_on_public_registry_risk": False,
    }
    info = pkg_metadata.get("info", {})
    versions_dict = pkg_metadata.get("releases", {})
    versions = [ver[0] for k, ver in versions_dict.items() if ver]
    is_deprecated = info.get("yanked") and info.get("yanked_reason")
    latest_deprecated = False
    first_version = None
    latest_version = None

    # Is the private package available publicly? Dependency confusion.
    if is_private_pkg and pkg_metadata:
        risk_metrics["pkg_private_on_public_registry_risk"] = True
        risk_metrics["pkg_private_on_public_registry_value"] = 1

    # If the package has fewer than minimum number of versions
    if len(versions):
        if len(versions) < config.pkg_min_versions:
            risk_metrics["pkg_min_versions_risk"] = True
            risk_metrics["pkg_min_versions_value"] = len(versions)
        first_version = versions[0]
        latest_version = versions[-1]
        # Check if the latest version is deprecated
        if latest_version and latest_version.get("yanked"):
            latest_deprecated = True

    # Created and modified time related checks
    if first_version and latest_version:
        created = first_version.get("upload_time")
        modified = latest_version.get("upload_time")
        if created and modified:
            modified_dt = datetime.fromisoformat(modified)
            created_dt = datetime.fromisoformat(created)
            mod_create_diff = modified_dt - created_dt
            latest_now_diff = datetime.now() - modified_dt
            created_now_diff = datetime.now() - created_dt
            risk_metrics = compute_time_risks(
                risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
            )

    # Is the package deprecated
    if is_deprecated or latest_deprecated:
        risk_metrics["pkg_deprecated_risk"] = True
        risk_metrics["pkg_deprecated_value"] = 1
    # Add package scope related weight
    if scope:
        risk_metrics[f"pkg_{scope}_scope_risk"] = True
        risk_metrics[f"pkg_{scope}_scope_value"] = 1

    risk_metrics["risk_score"] = calculate_risk_score(risk_metrics)
    return risk_metrics


def npm_pkg_risk(pkg_metadata, is_private_pkg, scope):
    """
    Calculate various npm package risks based on the metadata from npm. The
    keys in the risk_metrics dict is based on the parameters specified in
    config.py and has a _risk suffix. Eg: config.pkg_min_versions would
    result in a boolean pkg_min_versions_risk and pkg_min_versions_value

    :param pkg_metadata: A dict containing the metadata of the npm package.
    :param is_private_pkg: Boolean to indicate if this package is private
    :param scope: Package scope
    :return: A dict containing the calculated risks and score.
    """
    # Some default values to ensure the structure is non-empty
    risk_metrics = {
        "pkg_deprecated_risk": False,
        "pkg_min_versions_risk": False,
        "created_now_quarantine_seconds_risk": False,
        "latest_now_max_seconds_risk": False,
        "mod_create_min_seconds_risk": False,
        "pkg_min_maintainers_risk": False,
        "pkg_node_version_risk": False,
        "pkg_private_on_public_registry_risk": False,
    }
    # Is the private package available publicly? Dependency confusion.
    if is_private_pkg and pkg_metadata:
        risk_metrics["pkg_private_on_public_registry_risk"] = True
        risk_metrics["pkg_private_on_public_registry_value"] = 1
    versions = pkg_metadata.get("versions", {})
    latest_version = pkg_metadata.get("dist-tags", {}).get("latest")
    engines_block_dict = versions.get(latest_version, {}).get("engines", {})
    # Check for scripts block
    scripts_block_dict = versions.get(latest_version, {}).get("scripts", {})
    is_deprecated = versions.get(latest_version, {}).get("deprecated", None)
    # Is the package deprecated
    if is_deprecated:
        risk_metrics["pkg_deprecated_risk"] = True
        risk_metrics["pkg_deprecated_value"] = 1
    scripts_block_list = []
    # There are some packages on npm with incorrectly configured scripts
    # block Good news is that the install portion would only for if the
    # scripts block is an object/dict
    if isinstance(scripts_block_dict, dict):
        scripts_block_list = [
            block
            for block in scripts_block_dict.keys()
            if "preinstall" in block or "postinstall" in block
        ]

    # If the package has fewer than minimum number of versions
    if len(versions) < config.pkg_min_versions:
        risk_metrics["pkg_min_versions_risk"] = True
        risk_metrics["pkg_min_versions_value"] = len(versions)
    # Time related checks
    time_info = pkg_metadata.get("time", {})
    modified = time_info.get("modified", "").replace("Z", "")
    created = time_info.get("created", "").replace("Z", "")
    if not modified and pkg_metadata.get("mtime"):
        modified = pkg_metadata.get("mtime").replace("Z", "")
    if not created and pkg_metadata.get("ctime"):
        created = pkg_metadata.get("ctime").replace("Z", "")
    latest_version_time = time_info.get(latest_version, "").replace("Z", "")
    if time_info and modified and created and latest_version_time:
        modified_dt = datetime.fromisoformat(modified)
        created_dt = datetime.fromisoformat(created)
        latest_version_time_dt = datetime.fromisoformat(latest_version_time)
        mod_create_diff = modified_dt - created_dt
        latest_now_diff = datetime.now() - latest_version_time_dt
        created_now_diff = datetime.now() - created_dt
        risk_metrics = compute_time_risks(
            risk_metrics, created_now_diff, mod_create_diff, latest_now_diff
        )

    # Maintainers count related risk. Ignore packages that are past
    # quarantine period
    maintainers = pkg_metadata.get("maintainers", [])
    if len(maintainers) < config.pkg_min_maintainers and risk_metrics.get(
        "created_now_quarantine_seconds_risk"
    ):
        risk_metrics["pkg_min_maintainers_risk"] = True
        risk_metrics["pkg_min_maintainers_value"] = len(maintainers)
        # Check for install scripts risk only for those packages with
        # maintainers risk
        if scripts_block_list:
            risk_metrics["pkg_install_scripts_risk"] = True
            risk_metrics["pkg_install_scripts_value"] = len(scripts_block_list)

    # Users count related risk. Ignore packages that are past quarantine period
    users = pkg_metadata.get("users", [])
    if (
        users
        and len(users) < config.pkg_min_users
        and risk_metrics.get("created_now_quarantine_seconds_risk")
    ):
        risk_metrics["pkg_min_users_risk"] = True
        risk_metrics["pkg_min_users_value"] = len(users)
    # Node engine version There are packages with incorrect node engine
    # specification which we can ignore for now
    if (
        engines_block_dict
        and isinstance(engines_block_dict, dict)
        and engines_block_dict.get("node")
        and isinstance(engines_block_dict.get("node"), str)
    ):
        node_version_spec = engines_block_dict.get("node")
        node_version = (
            node_version_spec.replace(">= ", "")
            .replace(">=", "")
            .replace("> ", "")
            .replace(">", "")
            .replace("~ ", "")
            .replace("~", "")
            .split(" ")[0]
        )
        for ver in config.pkg_node_version.split(","):
            if node_version.startswith(ver):
                risk_metrics["pkg_node_version_risk"] = True
                risk_metrics["pkg_node_version_value"] = 1
                break
    # Add package scope related weight
    if scope:
        risk_metrics[f"pkg_{scope}_scope_risk"] = True
        risk_metrics[f"pkg_{scope}_scope_value"] = 1

    risk_metrics["risk_score"] = calculate_risk_score(risk_metrics)
    return risk_metrics



import json
import os
import re

from rich import box
from rich.markdown import Markdown
from rich.table import Table
from rich.tree import Tree

from depscan.lib.config import max_purl_per_flow, max_reachable_explanations
from depscan.lib.logger import console


def explain(
    project_type,
    src_dir,
    reachables_slices_file,
    vdr_file,
    pkg_vulnerabilities,
    pkg_group_rows,
    direct_purls,
    reached_purls,
):
    """
    Explain the analysis and findings

    :param project_type: Project type
    :param src_dir: Source directory
    :param reachables_slices_file: Reachables slices file
    :param vdr_file: VDR file from the summariser
    :param pkg_vulnerabilities: Vulnerabilities from the analysis
    :param pkg_group_rows: Prioritized list of purls
    :param direct_purls: Dict containing packages used directly
    :param reached_purls: Dict containing packages identified via reachables slicing
    """
    if (
        not reachables_slices_file
        and src_dir
        and os.path.exists(os.path.join(src_dir, "reachables.slices.json"))
    ):
        reachables_slices_file = os.path.join(src_dir, "reachables.slices.json")
    if reachables_slices_file:
        with open(reachables_slices_file, "r", encoding="utf-8") as f:
            reachables_data = json.load(f)
            if reachables_data:
                rsection = Markdown(
                    """## Reachable Flows

Below are some reachable flows identified by depscan. Use the provided tips to improve the securability of your application.
                """
                )
                console.print(rsection)
                explain_reachables(reachables_data, pkg_group_rows, project_type)


def explain_reachables(reachables, pkg_group_rows, project_type):
    """"""
    reachable_explanations = 0
    checked_flows = 0
    for areach in reachables.get("reachables", []):
        if (
            not areach.get("flows")
            or len(areach.get("flows")) < 2
            or not areach.get("purls")
        ):
            continue
        # Focus only on the prioritized list if available
        # if project_type in ("java",) and pkg_group_rows:
        #     is_prioritized = False
        #     for apurl in areach.get("purls"):
        #         if pkg_group_rows.get(apurl):
        #             is_prioritized = True
        #     if not is_prioritized:
        #         continue
        flow_tree, comment, source_sink_desc, has_check_tag = explain_flows(
            areach.get("flows"), areach.get("purls"), project_type
        )
        if not source_sink_desc:
            continue
        rtable = Table(
            box=box.DOUBLE_EDGE,
            show_lines=True,
            caption=comment,
            show_header=False,
            title=f"[bold]#{reachable_explanations + 1}[/bold] {source_sink_desc}",
            title_justify="left",
            min_width=150,
        )
        rtable.add_column(header="Flow", vertical="top")
        rtable.add_row(flow_tree)
        console.print()
        console.print(rtable)
        reachable_explanations += 1
        if has_check_tag:
            checked_flows += 1
        if reachable_explanations + 1 > max_reachable_explanations:
            break
    if reachable_explanations:
        tips = """## Secure Design Tips"""

        if checked_flows:
            tips += """
- Review the detected validation/sanitization methods in the application.
- To improve the security posture, implement a common validation middleware.
"""
        else:
            tips += """
- Consider implementing a common validation/sanitization library to reduce the exploitability risk.
"""
        rsection = Markdown(tips)
        console.print(rsection)


def flow_to_source_sink(flow, purls, project_type):
    """ """
    source_sink_desc = ""
    param_name = flow.get("name")
    method_str = "method"
    param_str = "Parameter"
    if param_name == "this":
        param_name = ""
    # Improve the labels based on the language
    if re.search(".(js|ts|jsx|tsx|py|cs)$", flow.get("parentFileName", "")):
        method_str = "function"
        param_str = "argument"
    if flow.get("label") == "METHOD_PARAMETER_IN":
        if param_name:
            source_sink_desc = f"""{param_str} [red]{param_name}[/red] :right_arrow_curving_left: to the {method_str} [bold]{flow.get('parentMethodName')}[/bold]"""
        else:
            source_sink_desc = f"""{method_str.capitalize()} [red]{flow.get('parentMethodName')}[/red] :right_arrow_curving_left:"""
    elif flow.get("label") == "CALL" and flow.get("isExternal"):
        method_full_name = flow.get("fullName", "")
        if not method_full_name.startswith("<"):
            source_sink_desc = f"Invocation: {method_full_name}"
    elif flow.get("label") == "RETURN" and flow.get("code"):
        source_sink_desc = flow.get("code").split("\n")[0]
    elif project_type not in ("java") and flow.get("label") == "IDENTIFIER":
        source_sink_desc = flow.get("code").split("\n")[0]
    # Try to understand the source a bit more
    if source_sink_desc.startswith("require("):
        source_sink_desc = "Flow starts from a module import"
    elif (
        ".use(" in source_sink_desc
        or ".subscribe(" in source_sink_desc
        or ".on(" in source_sink_desc
        or ".emit(" in source_sink_desc
        or " => {" in source_sink_desc
    ):
        source_sink_desc = "Flow starts from a callback function"
    elif (
        "middleware" in source_sink_desc.lower() or "route" in source_sink_desc.lower()
    ):
        source_sink_desc = "Flow starts from a middlware"
    elif len(purls) == 1:
        source_sink_desc = f"{source_sink_desc} can be used to reach this package."
    else:
        source_sink_desc = (
            f"{source_sink_desc} can be used to reach {len(purls)} packages."
        )
    return source_sink_desc


def filter_tags(tags):
    if tags:
        tags = [
            atag
            for atag in tags.split(", ")
            if atag not in ("RESOLVED_MEMBER", "UNKNOWN_METHOD", "UNKNOWN_TYPE_DECL")
        ]
        return ", ".join(tags)
    return tags


def flow_to_str(flow, project_type):
    """"""
    has_check_tag = False
    file_loc = ""
    if (
        flow.get("parentFileName")
        and flow.get("lineNumber")
        and not flow.get("parentFileName").startswith("unknown")
    ):
        file_loc = f'{flow.get("parentFileName").replace("src/main/java/", "").replace("src/main/scala/", "")}#{flow.get("lineNumber")}    '
    node_desc = flow.get("code").split("\n")[0]
    tags = filter_tags(flow.get("tags"))
    if flow.get("label") == "METHOD_PARAMETER_IN":
        param_name = flow.get("name")
        if param_name == "this":
            param_name = ""
        node_desc = f'{flow.get("parentMethodName")}([red]{param_name}[/red]) :right_arrow_curving_left:'
        if tags:
            node_desc = f"{node_desc}\n[bold]Tags:[/bold] [italic]{tags}[/italic]\n"
    elif flow.get("label") == "IDENTIFIER":
        if node_desc.startswith("<"):
            node_desc = flow.get("name")
        if tags:
            node_desc = f"{node_desc}\n[bold]Tags:[/bold] [italic]{tags}[/italic]\n"
    if tags:
        for ctag in (
            "validation",
            "encode",
            "encrypt",
            "sanitize",
            "authentication",
            "authorization",
        ):
            if ctag in tags:
                has_check_tag = True
                break
    if has_check_tag:
        node_desc = f"[green]{node_desc}[/green]"
    return file_loc, f"""[gray37]{file_loc}[/gray37]{node_desc}""", has_check_tag


def explain_flows(flows, purls, project_type):
    """"""
    tree = None
    comments = []
    if len(purls) > max_purl_per_flow:
        comments.append(
            ":exclamation_mark: Refactor this flow to reduce the number of external libraries used."
        )
    purls_str = "\n".join(purls)
    comments.append(f"[info]Reachable Packages:[/info]\n{purls_str}")
    added_flows = []
    has_check_tag = False
    last_file_loc = None
    source_sink_desc = ""
    for aflow in flows:
        # For java, we are only interested in identifiers with tags to keep the flows simple to understand
        if (
            project_type in ("java", "jar", "android", "apk")
            and aflow.get("label") == "IDENTIFIER"
            and not aflow.get("tags")
        ):
            continue
        if not source_sink_desc:
            source_sink_desc = flow_to_source_sink(aflow, purls, project_type)
        file_loc, flow_str, has_check_tag_flow = flow_to_str(aflow, project_type)
        if last_file_loc == file_loc:
            continue
        last_file_loc = file_loc
        if flow_str in added_flows:
            continue
        added_flows.append(flow_str)
        if not tree:
            tree = Tree(flow_str)
        else:
            tree.add(flow_str)
        if has_check_tag_flow:
            has_check_tag = True
    if has_check_tag:
        comments.insert(
            0,
            ":white_medium_small_square: Check if the mitigation(s) used in this flow is valid and appropriate for your security requirements.",
        )
    return tree, "\n".join(comments), source_sink_desc, has_check_tag



import os
import sys
from os.path import dirname, exists, join


def resource_path(relative_path):
    """

    :param relative_path:
    :return:
    """
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = dirname(__file__)
    return join(base_path, relative_path)


license_data_dir = resource_path(
    join(
        "..",
        "..",
        "vendor",
        "choosealicense.com",
        "_licenses",
    )
)
spdx_license_list = resource_path(
    join(
        "..",
        "..",
        "vendor",
        "spdx",
        "json",
        "licenses.json",
    )
)
if not exists(license_data_dir):
    license_data_dir = resource_path(
        join(
            "vendor",
            "choosealicense.com",
            "_licenses",
        )
    )
    spdx_license_list = resource_path(
        join(
            "vendor",
            "spdx",
            "json",
            "licenses.json",
        )
    )

# CPE Vendor aliases
vendor_alias = {
    "org.apache.commons.io": "commons-io",
    "org.apache.logging.log4j": "log4j",
    "org.apache.commons.beanutils": "commons-beanutils",
    "org.apache.commons.collections": "commons-collections",
    "org.apache.solr": "apache_solr",
    "org.springframework": "pivotal_software",
    "io.undertow": "redhat",
    "ch.qos.logback": "logback",
    "ch.qos.slf4j": "slf4j",
    "org.yaml": "snakeyaml_project",
    "org.hibernate.validator": "org.hibernate",
    "org.hibernate": "redhat",
    "org.dom4j": "dom4j_project",
    "ant": "apache",
    "commons-": "apache",
    "org.quartz-scheduler": "softwareag",
    "org.mitre": "mitreid",
    "io.micronaut": "objectcomputing",
    "twistedmatrix": "twisted",
    "oneup": "1up",
    "io.ktor": "jetbrains",
    "com.puppycrawl.tools": "checkstyle",
    "org.opencastproject": "apereo",
    "bagisto": "webkul",
    "ro.pippo": "pippo",
    "ca.uhn.hapi.fhir": "fhir",
    "tensorflow": "google",
    "ansible": "redhat",
    "io.springfox": "smartbear",
    "log4net": "apache",
    "github": "github actions",
}

# Package aliases
package_alias = {
    "struts2-core": "struts",
    "struts2-rest-plugin": "struts",
    "struts2-showcase": "struts",
    "jackson-databind": "jackson",
    "apache_tomcat": "tomcat",
    "tomcat_native": "tomcat",
    "tomcat_connectors": "tomcat",
    "tomcat_jk_connector": "tomcat",
    "tomcat-embed-core": "tomcat",
    "spring-security-core": "spring_security",
    "spring-security-crypto": "spring_security",
    "asciidoctorj": "asciidoctor",
    "postgresql": "postgresql_jdbc_driver",
    "itextpdf": "itext",
    "httpclient": "commons-httpclient",
    "priority": "python_priority_library",
    "rocketmq-broker": "rocketmq",
    "mysql_connector": "mysql-connector-java",
    "jhipster_kotlin": "jhipster",
    "spring-cloud-config-server": "spring_cloud_config",
    "django-rest-framework-json_web_tokens": "drf-jwt",
    "beam-sdks-java-io-mongodb": "beam",
    "sm-core-model": "shopizer",
    "openid-connect-server": "connect",
    "http4s-server_2.12": "http4s",
    "santuario_xml_security_for_java": "xmlsec",
    "uploader-bundle": "oneupuploaderbundle",
    "odata-client-core": "olingo",
    "odata-client-proxy": "olingo",
    "odata-server-core": "olingo",
    "syliusresourcebundle": "sylius",
    "ethereum_name_service": "ens",
    "tensorflow-gpu": "tensorflow",
    "tensorflow-cpu": "tensorflow",
    "class.upload.php": "verot",
    "redis_wrapper": "rediswrapper",
    "silverstripe-versionedfiles": "versionedfiles",
    "simplesamlphp-module-proxystatistics": "proxystatistics",
    "pac4j-saml": "pac4j",
    "universal_office_converter": "unoconv",
    "hapi-fhir-base": "hapi_fhir",
    "spring-data-jpa": "spring_data_java_persistance_api",
    "sanselan": "commons_imaging",
    "uima-ducc-web": "unstructured_information_management_architecture_distributed_uima_cluster_computing",
    "arrow-ank-gradle": "arrow",
    "openpgpjs": "openpgp",
    "storm-kafka": "storm",
    "storm-kafka-client": "storm",
    "tika-parsers": "tika",
    "ironic-discoverd": "ironic_inspector",
    "hawkbit-ui": "hawkbit",
    "hawkbit-starters": "hawkbit",
    "hawkbit-boot-starter": "hawkbit",
    "software_development_kit": "splunk-sdk",
    "jira_software_data_center": "jira",
    "springfox-swagger2": "swagger_ui",
    "spring-web": "spring_framework",
    "springfox-swagger-ui": "swagger_ui",
    "hibernate-core": "hibernate_orm",
    "json-smart": "json-smart-v2",
    "ojdbc7": "jdbc",
    "System.Text": ".net",
    "Microsoft.IdentityModel.Clients.ActiveDirectory": "active_directory_authentication_library",
    "starkbank_ecdsa": "ecdsa-elixir",
    "php-pear": "pear-core-minimal",
}

# Default ignore list
ignore_directories = [
    ".git",
    ".svn",
    ".mvn",
    ".idea",
    "dist",
    "bin",
    "obj",
    "backup",
    "docs",
    "tests",
    "test",
    "tmp",
    "report",
    "reports",
    "node_modules",
    ".terraform",
    ".serverless",
    "venv",
    "examples",
    "tutorials",
    "samples",
    "migrations",
    "db_migrations",
    "unittests",
    "unittests_legacy",
    "stubs",
    "mock",
    "mocks",
]

# Package types allowed for each language
LANG_PKG_TYPES = {
    "python": "pypi",
    "java": "maven",
    "jvm": "maven",
    "groovy": "maven",
    "kotlin": "maven",
    "scala": "maven",
    "js": "npm",
    "javascript": "npm",
    "nodejs": "npm",
    "go": "golang",
    "golang": "golang",
    "ruby": "gem",
    "php": "composer",
    "dotnet": "nuget",
    "csharp": "nuget",
    "rust": "cargo",
    "dart": "pub",
    "cpp": "conan",
    "clojure": "clojars",
    "haskell": "hackage",
    "elixir": "hex",
    "github actions": "github",
    "github": "github",
}

# OS Package types
OS_PKG_TYPES = (
    "deb",
    "apk",
    "rpm",
    "swid",
    "alpm",
    "docker",
    "oci",
    "container",
    "generic",
    "qpkg",
    "buildroot",
    "coreos",
    "ebuild",
    "alpine",
    "alma",
    "almalinux",
    "debian",
    "ubuntu",
    "amazon",
    "rhel",
    "redhat",
    "rocky",
    "arch",
    "suse",
    "photon",
    "microsoft",
)

# List of Linux distros with support for editions
LINUX_DISTRO_WITH_EDITIONS = (
    "debian",
    "ubuntu",
    "alpine",
    "rhel",
    "redhat",
    "arch",
    "suse",
    "photon",
    "alma",
    "almalinux",
    "amazon",
    "rocky",
)


def get_float_from_env(name, default):
    """
    Retrieves a value from an environment variable and converts it to a
    float. If the value cannot be converted to a float, it returns the
    default value provided.

    :param name:
    :param default:
    :return:
    """
    value = os.getenv(name.upper(), default)
    try:
        value = float(value)
    except ValueError:
        value = default
    return value


def get_int_from_env(name, default):
    """
    Retrieves a value from an environment variable and converts it to an
    integer. If the value cannot be converted to an integer, it returns the
    default value provided.

    :param name:
    :param default:
    """
    return int(get_float_from_env(name, default))


npm_server = "https://registry.npmjs.org"
npm_app_info = {"name": "appthreat-depscan", "version": "1.0.0"}

pypi_server = "https://pypi.org/pypi"

vdb_database_url = "ghcr.io/appthreat/vdb:v5"
vdb_rafs_database_url = "ghcr.io/appthreat/vdb:v5-rafs"

# Package risk scoring using a simple weighted formula with no backing
# research All parameters and their max value and weight can be overridden
# using environment variables

# Some constants and defaults
seconds_in_day = 24 * 60 * 60
seconds_in_hour = 60 * 60
default_max_value = 100
default_weight = 1

# Package should have at least 3 versions
pkg_min_versions = get_float_from_env("pkg_min_versions", 3)
pkg_min_versions_max = get_float_from_env("pkg_min_versions_max", 100)
pkg_min_versions_weight = get_float_from_env("pkg_min_versions_weight", 2)

# At least 12 hours difference between the creation and modified time
mod_create_min_seconds = get_float_from_env(
    "mod_create_min_seconds", 12 * seconds_in_hour
)
mod_create_min_seconds_max = get_float_from_env(
    "mod_create_min_seconds_max", 1000 * seconds_in_day
)
mod_create_min_seconds_weight = get_float_from_env("mod_create_min_seconds_weight", 1)

# At least 12 hours difference between the latest version and the current time
latest_now_min_seconds = get_float_from_env(
    "latest_now_min_seconds", 12 * seconds_in_hour
)
latest_now_min_seconds_max = get_float_from_env(
    "latest_now_min_seconds_max", 1000 * seconds_in_day
)
latest_now_min_seconds_weight = get_float_from_env("latest_now_min_seconds_weight", 0.5)

# Time period after which certain risks can be considered safe. Quarantine
# period For eg: Packages that are over 1 year old
created_now_quarantine_seconds = get_float_from_env(
    "created_now_quarantine_seconds", 365 * seconds_in_day
)
created_now_quarantine_seconds_max = get_float_from_env(
    "created_now_quarantine_seconds_max", 365 * seconds_in_day
)
created_now_quarantine_seconds_weight = get_float_from_env(
    "created_now_quarantine_seconds_weight", 0.5
)

# Max package age - 6 years
latest_now_max_seconds = get_float_from_env(
    "latest_now_max_seconds", 6 * 365 * seconds_in_day
)
latest_now_max_seconds_max = get_float_from_env(
    "latest_now_max_seconds_max", 6 * 365 * seconds_in_day
)
latest_now_max_seconds_weight = get_float_from_env("latest_now_max_seconds_weight", 0.5)

# Package should have at least 2 maintainers
pkg_min_maintainers = get_float_from_env("pkg_min_maintainers", 2)
pkg_min_maintainers_max = get_float_from_env("pkg_min_maintainers_max", 10)
pkg_min_maintainers_weight = get_float_from_env("pkg_min_maintainers_weight", 2)

# Package should have at least 2 users
pkg_min_users = get_float_from_env("pkg_min_users", 2)
pkg_min_users_max = get_float_from_env("pkg_min_users_max", 20)
pkg_min_users_weight = get_float_from_env("pkg_min_users_weight", 0.25)

# Package with install scripts (npm)
pkg_install_scripts_max = get_float_from_env("pkg_install_scripts_max", 0)
pkg_install_scripts_weight = get_float_from_env("pkg_install_scripts_weight", 2)

# Node version risk
pkg_node_version = os.getenv("pkg_node_version".upper(), "0.,4,6")
pkg_node_version_max = get_float_from_env("pkg_node_version_max", 16)
pkg_node_version_weight = get_float_from_env("pkg_node_version_weight", 0.5)

# Package deprecated
pkg_deprecated_weight = get_float_from_env("pkg_deprecated_weight", 1)
pkg_deprecated_max = get_float_from_env("pkg_deprecated_max", 1)

# Package dependency confusion
pkg_private_on_public_registry_weight = get_float_from_env(
    "pkg_private_on_public_registry_weight", 4
)
pkg_private_on_public_registry_max = get_float_from_env(
    "pkg_private_on_public_registry_max", 1
)

# Package scope related weight
pkg_required_scope_weight = get_float_from_env("pkg_required_scope_weight", 4.0)
pkg_optional_scope_weight = get_float_from_env("pkg_optional_scope_weight", 0.5)
pkg_excluded_scope_weight = get_float_from_env("pkg_excluded_scope_weight", 0)
pkg_required_scope_max = get_float_from_env("pkg_required_scope_max", 1)
pkg_optional_scope_max = get_float_from_env("pkg_optional_scope_max", 1)
pkg_excluded_scope_max = get_float_from_env("pkg_excluded_scope_max", 1)

total_weight = (
    pkg_min_versions_weight
    + mod_create_min_seconds_weight
    + latest_now_min_seconds_weight
    + latest_now_max_seconds_weight
    + created_now_quarantine_seconds_weight
    + pkg_min_maintainers_weight
    + pkg_min_users_weight
    + pkg_install_scripts_weight
    + pkg_node_version_weight
    + pkg_required_scope_weight
    + pkg_optional_scope_weight
    + pkg_deprecated_weight
    + pkg_private_on_public_registry_weight
)

# Help text for various risk
risk_help_text = {
    "pkg_min_versions": "Has fewer versions",
    "latest_now_min_seconds": "Recently updated",
    "latest_now_max_seconds": "No recent updates",
    "pkg_min_maintainers": "Has fewer maintainers",
    "pkg_node_version": "Outdated Node version",
    "pkg_install_scripts": "Runs scripts on install",
    "pkg_deprecated": "Deprecated",
    "pkg_private_on_public_registry": "Private package is public",
}

# Package max risk score. All packages above this level will be reported
pkg_max_risk_score = get_float_from_env("pkg_max_risk_score", 0.5)

# Default request timeout
request_timeout_sec = get_int_from_env("request_timeout_sec", 20)

# Number of api failures that would stop the risk audit completely
max_request_failures = get_int_from_env("max_request_failures", 5)

# Universal scan
UNIVERSAL_SCAN_TYPE = "universal"

max_reachable_explanations = get_int_from_env("max_reachable_explanations", 20)

max_purl_per_flow = get_int_from_env("max_purl_per_flow", 6)



from vdb.lib.npm import NpmSource

from depscan.lib import config
from depscan.lib.pkg_query import npm_metadata, pypi_metadata

# Dict mapping project type to the audit source
type_audit_map = {"nodejs": NpmSource(), "js": NpmSource(), "npm": NpmSource()}

# Dict mapping project type to risk audit
risk_audit_map = {
    "npm": npm_metadata,
    "nodejs": npm_metadata,
    "js": npm_metadata,
    "python": pypi_metadata,
    "py": pypi_metadata,
    "pypi": pypi_metadata,
}


def audit(project_type, pkg_list):
    """
    Method to audit packages using remote source such as npm advisory

    :param project_type: Project type
    :param pkg_list: List of packages
    :return: Results
    """
    results = type_audit_map[project_type].bulk_search(
        app_info=config.npm_app_info, pkg_list=pkg_list
    )
    return results


def risk_audit(project_type, scoped_pkgs, private_ns, pkg_list):
    """
    Method to perform risk audit for packages using package managers api

    :param scoped_pkgs: A list of scoped packages.
    :param project_type: Project type
    :param private_ns: Private namespace
    :param pkg_list: List of packages
    :return: Results of risk audit
    """
    audit_fn = risk_audit_map[project_type]
    results = audit_fn(scoped_pkgs, pkg_list, private_ns)
    return results



from depscan.lib.normalize import create_pkg_variations


def test_pkg_variations():
    pkg_list = create_pkg_variations(
        {"vendor": "fasterxml", "name": "jackson-databind", "version": "1.0.0"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "com.fasterxml.jackson.core",
            "name": "jackson-databind",
            "version": "1.0.0",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "commons-io", "name": "commons-io", "version": "1.0.0"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "org.eclipse.foo", "name": "bar", "version": "1.0.0"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "com.fasterxml.jackson.core",
            "name": "jackson-annotations",
            "version": "1.0.0",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "io.undertow", "name": "undertow-core", "version": "2.0.27.Final"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "io.undertow", "name": "undertow-core", "version": "2.0.27.Final"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "org.apache.logging.log4j", "name": "log4j-api", "version": "2.12.1"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "org.springframework.batch",
            "name": "spring-batch",
            "version": "2.0.27.Final",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "commons-fileupload",
            "name": "commons-fileupload",
            "version": "1.3.2",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {"vendor": "github.com/go-sql-driver", "name": "mysql", "version": "v1.4.1"}
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "golang.org/x/crypto",
            "name": "ssh",
            "version": "0.0.0-20200220183623-bac4c82f6975",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "github.com/mitchellh",
            "name": "cli",
            "version": "6.14.1",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "github.com/jacobsa",
            "name": "crypto",
            "version": "6.14.1",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "org.hibernate",
            "name": "hibernate-core",
            "version": "5.4.18.Final",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "org.springframework.security",
            "name": "spring-security-crypto",
            "version": "5.3.3.RELEASE",
        }
    )
    assert len(pkg_list) > 1
    pkg_list = create_pkg_variations(
        {
            "vendor": "deb",
            "name": "gnome-accessibility-themes",
            "version": "3.28-1ubuntu3",
        }
    )
    assert len(pkg_list) > 1
